# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API client library."""
import datetime
import getpass
import json
import logging

import altair
import pandas

from . import error
from . import resource


logger = logging.getLogger('timesketch_api.aggregation')


class Aggregation(resource.SketchResource):
    """Aggregation object.

    Attributes:
        aggregator_name: name of the aggregator class used to
            generate the aggregation.
        chart_color: the color of the chart.
        chart_type: the type of chart that will be generated
            from this aggregation object.
        type: the type of aggregation object.
        search_id: a search ID if the aggregation is tied to a specific
            saved search.
    """

    def __init__(self, sketch):
        self._created_at = ''
        self._parameters = {}
        self._updated_at = ''
        self._group_id = None
        self.aggregator_name = ''
        self.chart_color = ''
        self.chart_type = ''
        self.chart_title = ''
        self.search_id = None
        self.type = None
        resource_uri = 'sketches/{0:d}/aggregation/explore/'.format(sketch.id)
        super().__init__(
            sketch=sketch, resource_uri=resource_uri)

    @property
    def created_at(self):
        """Returns a timestamp when the aggregation was created."""
        return self._created_at

    def _get_aggregation_buckets(self, entry, name=''):
        """Yields all buckets from an aggregation result object.

        Args:
            entry: result dict from an aggregation request.
            name: name of aggregation results that contains
                all aggregation buckets.

        Yields:
            A dict with each aggregation bucket as well as
            the bucket_name.
        """
        if 'buckets' in entry:
            for bucket in entry.get('buckets', []):
                bucket['bucket_name'] = name
                yield bucket
        else:
            for key, value in iter(entry.items()):
                if not isinstance(value, dict):
                    continue
                for bucket in self._get_aggregation_buckets(
                        value, name=key):
                    yield bucket

    def _run_aggregator(
            self, aggregator_name, parameters, search_id=None, chart_type=None):
        """Run an aggregator class.

        Args:
            aggregator_name: the name of the aggregator class.
            parameters: a dict with the parameters for the aggregation class.
            search_id: an optional integer value with a primary key to a
                saved search.
            chart_type: string with the chart type.

        Returns:
            A dict with the aggregation results.
        """
        resource_url = '{0:s}/sketches/{1:d}/aggregation/explore/'.format(
            self.api.api_root, self._sketch.id)

        if chart_type is None and parameters.get('supported_charts'):
            chart_type = parameters.get('supported_charts')
            if isinstance(chart_type, (list, tuple)):
                chart_type = chart_type[0]

        if chart_type:
            self.chart_type = chart_type

        if search_id:
            self.search_id = search_id

        self.aggregator_name = aggregator_name
        self.chart_color = parameters.get('chart_color', '')
        self._parameters = parameters

        form_data = {
            'aggregator_name': aggregator_name,
            'aggregator_parameters': parameters,
            'chart_type': chart_type,
            'view_id': search_id,
        }

        response = self.api.session.post(resource_url, json=form_data)
        if not error.check_return_status(response, logger):
            error.error_message(
                response, message='Unable to query results', error=ValueError)

        return error.get_response_json(response, logger)

    # pylint: disable=arguments-differ
    def from_saved(self, aggregation_id):
        """Initialize the aggregation object from a saved aggregation.

        Args:
            aggregation_id: integer value for the stored
                aggregation (primary key).
        """
        resource_uri = 'sketches/{0:d}/aggregation/{1:d}/'.format(
            self._sketch.id, aggregation_id)
        self._resource_id = aggregation_id
        resource_data = self.api.fetch_resource_data(resource_uri)
        data = resource_data.get('objects', [None])[0]
        if not data:
            return

        self.aggregator_name = data.get('agg_type')
        self.type = 'stored'

        self._created_at = data.get('created_at', '')
        self._updated_at = data.get('updated_at', '')
        self._group_id = data.get('aggregationgroup_id')

        label_string = data.get('label_string', '')
        if label_string:
            self._labels = json.loads(label_string)
        else:
            self._labels = []

        chart_type = data.get('chart_type')
        param_string = data.get('parameters', '')
        if param_string:
            parameters = json.loads(param_string)
        else:
            parameters = {}

        self._parameters = parameters

        self._username = data.get('user', {}).get('username', 'System')
        self.resource_data = self._run_aggregator(
            aggregator_name=self.aggregator_name, parameters=parameters,
            chart_type=chart_type)

    # pylint: disable=arguments-differ
    def from_manual(self, aggregate_dsl, **kwargs):
        """Initialize the aggregation object by running an aggregation DSL.

        Args:
            aggregate_dsl: Elasticsearch aggregation query DSL string.
        """
        super().from_manual(**kwargs)
        resource_url = '{0:s}/sketches/{1:d}/aggregation/explore/'.format(
            self.api.api_root, self._sketch.id)

        self.aggregator_name = 'DSL'
        self._username = getpass.getuser()
        self.type = 'DSL'

        date_now = datetime.datetime.now(datetime.timezone.utc)
        self._created_at = date_now.isoformat()
        self._updated_at = date_now.isoformat()

        form_data = {
            'aggregation_dsl': aggregate_dsl,
        }

        response = self.api.session.post(resource_url, json=form_data)
        if not error.check_return_status(response, logger):
            error.error_message(
                response, message='Unable to query results', error=ValueError)

        self.resource_data = error.get_response_json(response, logger)

    def from_aggregator_run(
            self, aggregator_name, aggregator_parameters,
            search_id=None, chart_type=None):
        """Initialize the aggregation object by running an aggregator class.

        Args:
            aggregator_name: name of the aggregator class to run.
            aggregator_parameters: a dict with the parameters of the aggregator
                class.
            search_id: an optional integer value with a primary key to a saved
                search.
            chart_type: optional string with the chart type.
        """
        self.type = 'aggregator_run'
        self._parameters = aggregator_parameters
        self._username = getpass.getuser()

        date_now = datetime.datetime.now(datetime.timezone.utc)
        self._created_at = date_now.isoformat()
        self._updated_at = date_now.isoformat()

        self.resource_data = self._run_aggregator(
            aggregator_name, aggregator_parameters, search_id, chart_type)

    def lazyload_data(self, refresh_cache=False):
        """Load resource data once and cache the result.

        Args:
            refresh_cache: Boolean indicating if to update cache.

        Returns:
            Dictionary with resource data.
        """
        if self.resource_data and not refresh_cache:
            return self.resource_data

        # TODO: Implement a method to refresh cache.
        return self.resource_data

    @property
    def parameters(self):
        """Property that returns the parameters of the aggregation."""
        return self._parameters

    @property
    def is_part_of_group(self):
        """Property that returns whether an agg is part of a group or not."""
        if self._group_id is None:
            return False

        return bool(self._group_id)

    @property
    def title(self):
        """Property that returns the chart title of an aggregation."""
        if self.chart_title:
            return self.chart_title

        data = self.lazyload_data()
        meta = data.get('meta', {})
        self.chart_title = meta.get('vega_chart_title', '')

        return self.chart_title

    @title.setter
    def title(self, new_title):
        """Set the chart title of an aggregation."""
        self.chart_title = new_title

    @property
    def chart(self):
        """Property that returns an altair Vega-lite chart."""
        return self.generate_chart()

    @property
    def description(self):
        """Property that returns the description string."""
        data = self.resource_data
        meta = data.get('meta', {})
        return meta.get('description', '')

    @description.setter
    def description(self, description):
        """Set the description of an aggregation."""
        if 'meta' not in self.resource_data:
            return
        meta = self.resource_data.get('meta')
        meta['description'] = description

    @property
    def name(self):
        """Property that returns the name of the aggregation."""
        data = self.resource_data
        meta = data.get('meta', {})
        name = meta.get('name')
        if name:
            return name
        return self.aggregator_name

    @name.setter
    def name(self, name):
        """Set the name of the aggregation."""
        if 'meta' not in self.resource_data:
            return
        meta = self.resource_data.get('meta')
        meta['name'] = name

    def add_label(self, label):
        """Add a label to the aggregation.

        Args:
            label (str): string with the label information.
        """
        if label in self._labels:
            return
        self._labels.append(label)
        self.save()

    def to_dict(self):
        """Returns a dict."""
        entries = {}
        entry_index = 1
        data = self.lazyload_data()
        for entry in data.get('objects', []):
            for bucket in self._get_aggregation_buckets(entry):
                entries['entry_{0:d}'.format(entry_index)] = bucket
                entry_index += 1
        return entries

    def to_pandas(self):
        """Returns a pandas DataFrame."""
        panda_list = []
        data = self.lazyload_data()
        for entry in data.get('objects', []):
            for bucket in self._get_aggregation_buckets(entry):
                panda_list.append(bucket)
        return pandas.DataFrame(panda_list)

    @property
    def updated_at(self):
        """Returns a timestamp when the aggregation was last updated."""
        return self._updated_at

    def generate_chart(self):
        """Returns an altair Vega-lite chart."""
        if not self.chart_type:
            raise TypeError('Unable to generate chart, missing a chart type.')

        if not self._parameters.get('supported_charts'):
            self._parameters['supported_charts'] = self.chart_type

        data = self.lazyload_data()
        meta = data.get('meta', {})
        vega_spec = meta.get('vega_spec')

        if not vega_spec:
            return altair.Chart(pandas.DataFrame()).mark_point()

        vega_spec_string = json.dumps(vega_spec)
        return altair.Chart.from_json(vega_spec_string)

    def save(self):
        """Save the aggregation in the database."""
        data = {
            'name': self.name,
            'description': self.description,
            'agg_type': self.aggregator_name,
            'parameters': self._parameters,
            'chart_type': self.chart_type,
        }
        if self.search_id:
            data['view_id'] = self.search_id
        if self._labels:
            data['labels'] = json.dumps(self._labels)

        if self._resource_id:
            resource_url = '{0:s}/sketches/{1:d}/aggregation/{2:d}/'.format(
                self.api.api_root, self._sketch.id, self._resource_id)
        else:
            resource_url = '{0:s}/sketches/{1:d}/aggregation/'.format(
                self.api.api_root, self._sketch.id)

        response = self.api.session.post(resource_url, json=data)
        if not error.check_return_status(response, logger):
            return 'Unable to save the aggregation'

        response_json = response.json()
        objects = response_json.get('objects')
        if not objects:
            return 'Unable to determine ID of saved object.'
        agg_data = objects[0]
        self._resource_id = agg_data.get('id', 0)
        return 'Saved aggregation to ID: {0:d}'.format(self._resource_id)

    def delete(self):
        """Deletes the aggregation from the store."""
        if not self._resource_id:
            logger.warning(
                'Unable to delete the aggregation, it does not appear to be '
                'saved in the first place.')
            return False

        resource_uri = '{0:s}/sketches/{1:d}/aggregation/{2:d}/'.format(
            self.api.api_root, self._sketch.id, self._resource_id)
        response = self.api.session.delete(resource_uri)
        return error.check_return_status(response, logger)


class AggregationGroup(resource.SketchResource):
    """Aggregation Group object."""

    def __init__(self, sketch):
        """Initialize the aggregation group."""
        resource_uri = 'sketches/{0:d}/aggregation/group/'.format(
            sketch.id)
        super().__init__(resource_uri=resource_uri, sketch=sketch)

        self._name = 'N/A'
        self._created_at = ''
        self._description = 'N/A'
        self._orientation = ''
        self._parameters = {}
        self._aggregations = []
        self._updated_at = ''

    def __str__(self):
        """Return a string representation of the group."""
        return '[{0:d}] {1:s} - {2:s}'.format(
            self._resource_id, self._name, self._description)

    @property
    def aggregations(self):
        """Property that returns a list of aggregations in the group."""
        return self._aggregations

    @property
    def created_at(self):
        """Returns a timestamp when the aggregation group was created."""
        return self._created_at

    @property
    def updated_at(self):
        """Returns a timestamp when the aggregation group was updated."""
        return self._updated_at

    def to_dict(self):
        """Returns the aggregation values as a dict."""
        data_frame = self.to_pandas()
        return data_frame.to_dict(orient='records')

    @property
    def chart(self):
        """Property that returns an altair Vega-lite chart."""
        if not self._aggregations:
            return altair.Chart()
        return self.generate_chart()

    @property
    def description(self):
        """Returns the description of the aggregation group."""
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of the aggregation group."""
        self._description = description
        self.save()

    @property
    def name(self):
        """Returns the name of the aggregation group."""
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of the aggregation group."""
        self._name = name
        self.save()

    @property
    def orientation(self):
        """Returns the chart orientation."""
        return self._orientation

    @orientation.setter
    def orientation(self, orientation):
        """Sets the chart orientation."""
        self._orientation = orientation
        self.save()

    @property
    def parameters(self):
        """Returns a dict with the group parameters."""
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """Sets the group parameters."""
        self._parameters = parameters
        self.save()

    def delete(self):
        """Deletes the group from the store."""
        if not self._resource_id:
            logger.warning(
                'Unable to delete the group, it does not appear to be saved '
                'in the first place.')
            return False

        resource_uri = '{0:s}/sketches/{1:d}/aggregation/group/{2:d}/'.format(
            self.api.api_root, self._sketch.id, self._resource_id)
        response = self.api.session.delete(resource_uri)
        return error.check_return_status(response, logger)

    def from_dict(self, group_dict):
        """Feed group data from a dictionary.

        Args:
            group_dict (dict): a dictionary with the aggregation group
                information.

        Raises:
            TypeError: if the dictionary does not contain the correct
                information.
        """
        group_id = group_dict.get('id')
        if not group_id:
            raise TypeError('Group ID is missing.')
        self._resource_id = group_id
        self.resource_uri = 'sketches/{0:d}/aggregation/group/{1:d}/'.format(
            self._sketch.id, group_id)

        self._name = group_dict.get('name', '')
        self._description = group_dict.get('description', '')

        self._created_at = group_dict.get('created_at', '')
        self._updated_at = group_dict.get('updated_at', '')

        self._orientation = group_dict.get('orientation')
        if not self._orientation:
            raise TypeError('How a group is connected needs to be defined.')

        parameter_string = group_dict.get('parameters', '')
        if parameter_string:
            self._parameters = json.loads(parameter_string)

        aggs = group_dict.get('agg_ids')
        if not aggs:
            raise TypeError('Group contains no aggregations')

        if isinstance(aggs, str):
            aggs = json.loads(aggs)

        if not isinstance(aggs, (list, tuple)):
            raise TypeError('Aggregations need to be a list.')

        self._aggregations = []
        for agg_id in aggs:
            agg_obj = Aggregation(self._sketch)
            agg_obj.from_saved(agg_id)
            self._aggregations.append(agg_obj)

    # pylint: disable=arguments-differ
    def from_saved(self, group_id):
        """Feed group data from a group ID.

        Args:
            group_id (int): the group ID to fetch from the store.

        Raises:
            TypeError: if the group ID does not exist.
        """
        self._resource_id = group_id
        resource_uri = (
            f'sketches/{self._sketch.id}/aggregation/group/{group_id}/')
        resource_data = self.api.fetch_resource_data(resource_uri)

        objects = resource_data.get('objects')
        if not objects:
            raise TypeError('Group ID not found.')

        group_dict = objects[0]
        group_dict['id'] = group_id
        self.from_dict(group_dict)

    def generate_chart(self):
        """Returns an altair Vega-lite chart."""
        if not self._aggregations:
            return altair.Chart()

        data = self.lazyload_data()

        meta = data.get('meta', {})
        vega_spec = meta.get('vega_spec')

        if not vega_spec:
            return altair.Chart(pandas.DataFrame()).mark_point()

        vega_spec_string = json.dumps(vega_spec)
        return altair.Chart.from_json(vega_spec_string)

    def get_charts(self):
        """Returns a list of altair Chart objects from each aggregation."""
        return [x.chart for x in self._aggregations]

    def get_tables(self):
        """Returns a list of pandas DataFrame from each aggregation."""
        return [x.table for x in self._aggregations]

    def save(self):
        """Save the aggregation group in the database."""
        if not self._aggregations:
            return False

        data = {
            'name': self._name,
            'description': self._description,
            'parameters': json.dumps(self._parameters),
            'aggregations': json.dumps([x.id for x in self._aggregations]),
            'orientation': self._orientation,
        }

        if self._resource_id:
            res_url = '{0:s}/sketches/{1:d}/aggregation/group/{2:d}/'.format(
                self.api.api_root, self._sketch.id, self._resource_id)
        else:
            res_url = '{0:s}/sketches/{1:d}/aggregation/group/'.format(
                self.api.api_root, self._sketch.id)

        response = self.api.session.post(res_url, json=data)
        _ = self.lazyload_data(refresh_cache=True)
        return error.check_return_status(response, logger)

    def to_pandas(self):
        """Returns a pandas DataFrame.

        Aggregation groups are meant for charts, not data frames. However
        there are situations where you may want to be able to produce
        a DataFrame, and in that case all DataFrame objects from each
        aggregation are concatenated to return a single one.

        Returns:
            A single DataFrame that consists of a DataFrame from each
            aggregation in the group concatenated.
        """
        if not self._aggregations:
            return pandas.DataFrame()

        data_frames = []
        for agg_obj in self._aggregations:
            data_frames.append(agg_obj.to_pandas())

        return pandas.concat(data_frames)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API crypto storage library for OAUTH client."""
from __future__ import unicode_literals

import base64
import os
import getpass
import logging
import stat

from cryptography import fernet
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from . import credentials


logger = logging.getLogger('timesketch_api.crypto_client')


class CredentialStorage:
    """Class to store and retrieve stored credentials."""

    # The default filename of the token file.
    DEFAULT_CREDENTIAL_FILENAME = '.timesketch.token'

    # Length of the salt.
    SALT_LENGTH = 16

    def __init__(self, file_path=''):
        """Initialize the class."""
        self._user = getpass.getuser()
        if file_path:
            self._filepath = file_path
        else:
            home_path = os.path.expanduser('~')
            self._filepath = os.path.join(
                home_path, self.DEFAULT_CREDENTIAL_FILENAME)

    def _get_key(self, salt, password):
        """Returns an encryption key.

        Args:
            salt (bytes): a salt used during the encryption.
            password (bytes): the password used to decrypt/encrypt
                the message.

        Returns:
            Bytes with the encryption key.
        """
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
            backend=backends.default_backend()
        )
        return base64.urlsafe_b64encode(
            kdf.derive(password))

    def set_filepath(self, file_path):
        """Set the filepath to the credential file."""
        if os.path.isfile(file_path):
            self._filepath = file_path

    def save_credentials(
            self, cred_obj, file_path='', password='', config_assistant=None):
        """Save credential data to a token file.

        This function will create an encrypted file
        in the user's home directory (~/.timesketch.token by default)
        that contains a stored copy of the credential object.

        Args:
            cred_obj (credentials.TimesketchCredentials): the credential
                object that is to be stored on disk.
            file_path (str): full path to the file storing the saved
                credentials.
            password (str): optional password to encrypt the
                credential file with. If not supplied a password
                will be generated.
            config_assistant (ConfigAssistant): optional configuration
                assistant object. Can be used to store the password to the
                credential file.
        """
        if not file_path:
            file_path = self._filepath

        if password:
            password = bytes(password, 'utf-8')
        else:
            password = fernet.Fernet.generate_key()
            if config_assistant:
                config_assistant.set_config('cred_key', password)
                config_assistant.save_config()

        if not os.path.isfile(file_path):
            logger.info('File does not exist, creating it.')

        salt = os.urandom(self.SALT_LENGTH)
        key = self._get_key(salt, password)
        crypto = fernet.Fernet(key)

        data = cred_obj.serialize()

        with open(file_path, 'wb') as fw:
            fw.write(salt)
            fw.write(crypto.encrypt(data))

        file_permission = stat.S_IREAD | stat.S_IWRITE
        os.chmod(file_path, file_permission)
        logger.info('Credentials saved to: %s', file_path)

    def load_credentials(
            self, file_path='', password='', config_assistant=None):
        """Load credentials from a file and return a credential object.

        Args:
            file_path (str): Full path to the file storing the saved
                credentials.
            password (str): optional password to encrypt the
                credential file with. If not supplied a password
                will be generated.
            config_assistant (ConfigAssistant): optional configuration
                assistant object. Can be used to store the password to the
                credential file.

        Raises:
            IOError: If not able to decrypt the data.

        Returns:
            Credential object (oauth2.credentials.Credentials) read from
            the file.
        """
        if not file_path:
            file_path = self._filepath

        if not os.path.isfile(file_path):
            return None

        if password:
            password = bytes(password, 'utf-8')
        elif config_assistant:
            try:
                password = config_assistant.get_config('cred_key')
                if not isinstance(password, bytes):
                    password = bytes(password, 'utf-8')
            except KeyError as exc:
                raise IOError(
                    'Not able to determine encryption key from '
                    'config.') from exc
        else:
            raise IOError(
                'Neither password nor a configuration assistant passed to '
                'tool, unable to determine password.')

        with open(file_path, 'rb') as fh:
            salt = fh.read(self.SALT_LENGTH)
            key = self._get_key(salt, password)
            data = fh.read()
            crypto = fernet.Fernet(key)
            try:
                data_string = crypto.decrypt(data)
            except fernet.InvalidSignature as e:
                raise IOError(
                    'Unable to decrypt data, signature is not correct: '
                    '{0!s}'.format(e)) from e
            except fernet.InvalidToken as e:
                raise IOError(
                    'Unable to decrypt data, password wrong? (error '
                    '{0!s})'.format(e)) from e

            # TODO: Implement a manager.
            cred_obj = credentials.TimesketchPwdCredentials()
            try:
                cred_obj.deserialize(data_string)

                return cred_obj
            except TypeError:
                logger.debug('Credential object is not "timesketch" auth.')

            cred_obj = credentials.TimesketchOAuthCredentials()
            cred_obj.deserialize(data_string)
            return cred_obj



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API library for reading/parsing configs."""
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Text

import configparser
import logging
import os
import requests

from google.auth.transport import requests as auth_requests

from . import client
from . import cli_input
from . import credentials as ts_credentials
from . import crypto


logger = logging.getLogger('timesketch_api.config_assistance')


class ConfigAssistant:
    """A simple assistant that helps with setting up a Timesketch API client.

    This assistant can read and save configs to a file. It also maintains
    a state about what config has been passed to it, to assist possible
    UIs to understand what is missing in order to setup a Timesketch API
    client.
    """

    # The name of the default config file.
    RC_FILENAME = '.timesketchrc'

    # The needed items to configure a client.
    CLIENT_NEEDED = frozenset([
        'host_uri',
        'username',
        'auth_mode',
    ])
    OAUTH_CLIENT_NEEDED = frozenset([
        'client_id',
        'client_secret',
    ])

    EXTRA_FIELDS = frozenset([
        'token_file_path',
        'verify',
    ])

    CONFIG_ORDERING = {
        'host_uri': 1,
        'auth_mode': 2,
        'username': 3,
        'client_id': 4,
        'client_secret': 5,
        'password': 6,
    }

    CONFIG_HINTS = {
        'host_uri': 'URL of the Timesketch server',
        'username': 'The username of the Timesketch user',
        'password': 'Password of the chosen user.',
        'auth_mode': (
            'Authentication mode, valid choices are: "userpass" '
            '(user/pass) or "oauth"'),
        'client_id': 'OAUTH Client identification.',
        'client_secret': 'The OAUTH client secret',
    }

    def __init__(self):
        """Initialize the configuration assistant."""
        self._config = {}

    @property
    def parameters(self) -> List[Text]:
        """Return a list of configured parameters."""
        return self._config.keys()

    @property
    def missing(self) -> List[Text]:
        """Return a list of missing parameters."""
        return self.get_missing_config()

    def get_config(self, name: Text) -> Any:
        """Returns a value for a given config.

        Args:
            name (str): the name of the config value to retrieve.

        Raises:
            KeyError: if the config does not exist.
        """
        return self._config[name]

    def get_client(
            self, token_password: Optional[Text] = '') -> Optional[
                client.TimesketchApi]:
        """Returns a Timesketch API client if possible.

        Args:
            token_password (str): an optional password to decrypt
                the credential token file.
        """
        if self.missing:
            return None

        auth_mode = self._config.get('auth_mode', 'userpass')
        # TODO: Remove shortly, temporary due to change of
        # 'timesketch' to 'userpass'
        if auth_mode == 'timesketch':
            auth_mode = 'userpass'

        file_path = self._config.get('token_file_path', '')
        credential_storage = crypto.CredentialStorage(file_path=file_path)
        credentials = credential_storage.load_credentials(
            config_assistant=self, password=token_password)

        if auth_mode.startswith('oauth'):
            if not credentials:
                return client.TimesketchApi(
                    host_uri=self._config.get('host_uri'),
                    username=self._config.get('username'),
                    password=self._config.get('password', ''),
                    verify=self._config.get('verify', True),
                    client_id=self._config.get('client_id', ''),
                    client_secret=self._config.get('client_secret', ''),
                    auth_mode=auth_mode,
                )

            ts = client.TimesketchApi(
                host_uri=self._config.get('host_uri'),
                username=self._config.get('username'),
                auth_mode=auth_mode,
                create_session=False)
            ts.set_credentials(credentials)
            session = auth_requests.AuthorizedSession(credentials.credential)
            try:
                ts.refresh_oauth_token()
            except auth_requests.RefreshError as e:
                logger.error(
                    'Unable to refresh credentials, with error: %s', e)
                return None
            session = ts.authenticate_oauth_session(session)
            ts.set_session(session)
            return ts

        if credentials:
            username = credentials.credential.get(
                'username', self._config.get('username', ''))
            password = credentials.credential.get(
                'password', self._config.get('password', ''))
        else:
            username = self._config.get('username', '')
            password = self._config.get('password', '')

        return client.TimesketchApi(
            host_uri=self._config.get('host_uri'),
            username=username,
            password=password,
            verify=self._config.get('verify', True),
            client_id=self._config.get('client_id', ''),
            client_secret=self._config.get('client_secret', ''),
            auth_mode=auth_mode,
        )

    def get_missing_config(self) -> List[Text]:
        """Returns a list of configuration parameters that are missing.

        Returns:
            A list of parameters that are missing from the config object.
        """
        needed_set = self.CLIENT_NEEDED
        auth_mode = self._config.get('auth_mode', '')
        if auth_mode.startswith('oauth'):
            needed_set = needed_set.union(self.OAUTH_CLIENT_NEEDED)
        configured_set = set(self._config.keys())
        return sorted(
            list(needed_set.difference(configured_set)),
            key=lambda x: self.CONFIG_ORDERING.get(x, 100))

    def has_config(self, name: Text) -> bool:
        """Returns a boolean indicating whether a config parameter is set.

        Args:
            name (str): the name of the configuration.

        Returns:
            bool: whether the object has been set or not.
        """
        return name.lower() in self._config

    def load_config_file(
            self, config_file_path: Optional[Text] = '',
            section: Optional[Text] = 'timesketch',
            load_cli_config: Optional[bool] = False):
        """Load the config from file.

        Args:
            config_file_path (str): Full path to the configuration file,
                if not supplied the default path will be used, which is
                the file RC_FILENAME inside the user's home directory.
            section (str): The configuration section to read from. This
                is optional and defaults to timesketch. This can be
                useful if you have multiple Timesketch servers to connect to,
                with each one of them having a separate section in the config
                file.
            load_cli_config (bool): Determine if the CLI config section should
                be loaded. This is optional and defaults to False.

        Raises:
          IOError if the file does not exist or config does not load.
        """
        if config_file_path:
            if not os.path.isfile(config_file_path):
                error_msg = (
                    'Unable to load config file, file {0:s} does not '
                    'exist.').format(config_file_path)
                logger.error(error_msg)
                raise IOError(error_msg)
        else:
            home_path = os.path.expanduser('~')
            config_file_path = os.path.join(home_path, self.RC_FILENAME)

        if not os.path.isfile(config_file_path):
            fw = open(config_file_path, 'a')
            fw.close()

        config = configparser.ConfigParser()
        try:
            files_read = config.read([config_file_path])
        except configparser.MissingSectionHeaderError as exc:
            raise IOError(
                'Unable to parse config file') from exc

        if not files_read:
            logger.warning('No config read')
            return

        if not section:
            section = 'timesketch'

        if section not in config.sections():
            logger.warning('No %s section in the config', section)
            return

        timesketch_config = config[section]
        for name, value in timesketch_config.items():
            self.set_config(name, value)

        if load_cli_config:
            if 'cli' not in config.sections():
                # Set default CLI config section
                config['cli'] = {
                    'sketch': '',
                    'output_format': 'tabular'
                }

            cli_config = config['cli']
            for name, value in cli_config.items():
                self.set_config(name, value)

    def load_config_dict(self, config_dict: Dict[Text, Text]):
        """Loads configuration from a dictionary.

        Only loads the config items that are needed for the client,
        other keys are ignored in the dict object.

        Args:
            config_dict (dict): dict object with configuration.
        """
        fields = list(self.CLIENT_NEEDED)
        fields.extend(list(self.OAUTH_CLIENT_NEEDED))
        fields.extend(list(self.EXTRA_FIELDS))

        for key, value in config_dict.items():
            key = key.lower()
            if key not in fields:
                continue

            if not value:
                continue
            self.set_config(key, value)

    def save_config(
            self, file_path: Optional[Text] = '',
            section: Optional[Text] = 'timesketch',
            token_file_path: Optional[Text] = ''):
        """Save the current config to a file.

        Args:
            file_path (str): A full path to the location where the
                configuration file is to be stored. If not provided the
                default location will be used.
            section (str): The configuration section to write to. This
                is optional and defaults to timesketch. This can be
                useful if you have multiple Timesketch servers to connect to,
                with each one of them having a separate section in the config
                file.
            token_file_path (str): Optional path to the location of the token
                file.
        """
        if not file_path:
            home_path = os.path.expanduser('~')
            file_path = os.path.join(home_path, self.RC_FILENAME)

        config = configparser.ConfigParser()

        if not os.path.isfile(file_path):
            fw = open(file_path, 'a')
            fw.close()

        # Read in other sections in the config file as well.
        try:
            _ = config.read([file_path])
        except configparser.MissingSectionHeaderError:
            pass

        # TODO: Remove this, temporary here to transition from the use of
        # timesketch to the auth mode of userpass.
        auth_mode = self._config.get('auth_mode', 'userpass')
        if auth_mode == 'timesketch':
            auth_mode = 'userpass'

        if not section:
            section = 'timesketch'

        config[section] = {
            'host_uri': self._config.get('host_uri'),
            'username': self._config.get('username'),
            'verify': self._config.get('verify', True),
            'client_id': self._config.get('client_id', ''),
            'client_secret': self._config.get('client_secret', ''),
            'auth_mode': auth_mode,
        }
        config['cli'] = {
            'sketch': self._config.get('sketch', ''),
            'output_format': self._config.get('output_format', 'tabular')
        }
        if token_file_path:
            config[section]['token_file_path'] = token_file_path

        if 'cred_key' in self._config:
            cred_key = self._config.get('cred_key')
            if isinstance(cred_key, bytes):
                cred_key = cred_key.decode('utf-8')
            config[section]['cred_key'] = cred_key

        with open(file_path, 'w') as fw:
            config.write(fw)

    def set_config(self, name: Text, value: Any):
        """Sets a given config item with a value.

        Args:
          name (str): the name of the configuration value to be set.
          value (object): the value of the configuration object.
        """
        self._config[name.lower()] = value


def get_client(
        config_dict: Optional[Dict[Text, Any]] = None,
        config_path: Optional[Text] = '',
        config_section: Optional[Text] = 'timesketch',
        token_password: Optional[Text] = '',
        confirm_choices: Optional[bool] = False,
        load_cli_config: Optional[bool] = False
        ) -> Optional[client.TimesketchApi]:
    """Returns a Timesketch API client using the configuration assistant.

    Args:
        config_dict (dict): optional dict that will be used to configure
            the client.
        config_path (str): optional path to the configuration file, if
            not supplied a default path will be used.
        config_section (str): The configuration section to read from. This
            is optional and defaults to timesketch. This can be
            useful if you have multiple Timesketch servers to connect to,
            with each one of them having a separate section in the config
            file.
        token_password (str): an optional password to decrypt
            the credential token file.
        confirm_choices (bool): an optional bool. if set to the user is given
            a choice to change the value for all already configured parameters.
            This defaults to False.
        load_cli_config (bool): Determine if the CLI config section should
            be loaded. This is optional and defaults to False.

    Returns:
        A timesketch client (TimesketchApi) or None if not possible.
    """
    assistant = ConfigAssistant()
    try:
        assistant.load_config_file(
            config_path, section=config_section,
            load_cli_config=load_cli_config)
        if config_dict:
            assistant.load_config_dict(config_dict)
    except IOError as e:
        logger.error('Unable to load the config file, is it valid?')
        logger.error('Error: %s', e)

    try:
        configure_missing_parameters(
            config_assistant=assistant,
            token_password=token_password,
            confirm_choices=confirm_choices,
            config_section=config_section)
        return assistant.get_client(token_password=token_password)
    except (RuntimeError, requests.ConnectionError) as e:
        logger.error(
            'Unable to connect to the Timesketch server, are you '
            'connected to the network? Is the timesketch server '
            'running and accessible from your host? The error '
            'message is %s', e)
    except IOError as e:
        logger.error('Unable to get a client, with error: %s', e)
        logger.error(
            'If the issue is in the credentials then one solution '
            'is to remove the ~/.timesketch.token file and the '
            'credential section in ~/.timesketchrc or to remove '
            'both files. Or you could have supplied a wrong '
            'password to undecrypt the token file.')


def configure_missing_parameters(
        config_assistant: ConfigAssistant,
        token_password: Optional[Text] = '',
        confirm_choices: Optional[bool] = False,
        config_section: Optional[Text] = 'timesketch'):
    """Fill in missing configuration for a config assistant.

    This function will take in a configuration assistant object and check
    whether it is not fully configured. If it isn't it will ask the user
    to fill in the missing details.

    It will also check to see whether a password has been set if the auth
    is username/password and ask for a password to store credentials.

    Args:
        config_assistant (ConfigAssistant): a config assistant that might
            not be fully configured.
        token_password (str): an optional password to decrypt
            the credential token file.
        confirm_choices (bool): an optional bool. if set to the user is given
            a choice to change the value for all already configured parameters.
            This defaults to False.
        config_section (str): The configuration section to read from. This
            is optional and defaults to timesketch. This can be
            useful if you have multiple Timesketch servers to connect to,
            with each one of them having a separate section in the config
            file.
    """
    just_configured = []

    for field in config_assistant.missing:
        hint = config_assistant.CONFIG_HINTS.get(field, '')
        value = cli_input.ask_question(
            f'What is the value for <{field}> ({hint})', input_type=str)
        if value:
            config_assistant.set_config(field, value)
            just_configured.append(field)

    if config_assistant.missing:
        # We still have unanswered questions.
        return configure_missing_parameters(config_assistant, token_password)

    if confirm_choices:
        # Go through prior answered parameters.
        for field in config_assistant.parameters:
            if field in ('cred_key', 'verify', 'username'):
                continue
            # We don't want to re-ask the user about the field they just
            # configured.
            if field in just_configured:
                continue

            answer = config_assistant.get_config(field)
            change = cli_input.confirm_choice(
                f'Want to change the value for "{field}" [{answer}]',
                default=False, abort=False)
            if not change:
                continue
            hint = config_assistant.CONFIG_HINTS.get(field, '')
            value = cli_input.ask_question(
                f'What is the value for <{field}> ({hint})', input_type=str)
            if value:
                config_assistant.set_config(field, value)

    try:
        file_path = config_assistant.get_config('token_file_path')
    except KeyError:
        file_path = ''
    config_assistant.save_config(
        section=config_section, token_file_path=file_path)
    credential_storage = crypto.CredentialStorage(file_path=file_path)
    credentials = credential_storage.load_credentials(
        config_assistant=config_assistant, password=token_password)

    # Check if we are using username/password and we don't have credentials
    # saved.
    auth_mode = config_assistant.get_config('auth_mode')
    if auth_mode != 'userpass':
        return None

    choice = False
    if credentials and confirm_choices:
        choice = cli_input.confirm_choice(
            'Want to change credentials?',
            default=False, abort=False)

    if not choice and credentials:
        return None

    username = config_assistant.get_config('username')
    if choice:
        value = cli_input.ask_question(
            'What is the username?',
            input_type=str, default=username)
        if value:
            username = value
            config_assistant.set_config('username', value)

    password = cli_input.ask_question(
        'Password for user {0:s}'.format(username), input_type=str,
        hide_input=True, default='***')
    credentials = ts_credentials.TimesketchPwdCredentials()
    credentials.credential = {
        'username': username,
        'password': password
    }
    cred_storage = crypto.CredentialStorage(file_path=file_path)
    cred_storage.save_credentials(
        credentials, password=token_password,
        config_assistant=config_assistant)
    config_assistant.save_config(
        section=config_section, token_file_path=file_path)
    return None



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API graph object."""
import copy
import datetime
import json
import logging

import dateutil.parser
import pandas
import numpy
import networkx as nx

from . import error
from . import resource


logger = logging.getLogger('timesketch_api.graph')


class Graph(resource.SketchResource):
    """Graph object."""

    # This defines a list of layouts that are available
    # for the graph.
    _GRAPH_LAYOUTS = {
        'bipartite': nx.bipartite_layout,
        'circular': nx.circular_layout,
        'kamada': nx.kamada_kawai_layout,
        'kamada_kawai': nx.kamada_kawai_layout,
        'planar': nx.planar_layout,
        'random': nx.random_layout,
        'rescale': nx.rescale_layout,
        'shell': nx.shell_layout,
        'spectral': nx.spectral_layout,
        'multipartite': nx.multipartite_layout,
        'spring': nx.spring_layout,
    }

    def __init__(self, sketch):
        """Initialize the graph object."""
        resource_uri = f'sketches/{sketch.id}/graphs/'
        super().__init__(sketch=sketch, resource_uri=resource_uri)

        self._created_at = None
        self._graph = None
        self._description = ''
        self._layout = None
        self._name = ''
        self._timelines = []
        self._updated_at = None
        self._graph_config = {}

    def _parse_graph_dict(self, graph_dict):
        """Takes a dict object and constructs graph data from it.

        Args:
            graph_dict (dict): a dict that contains graph elements, as
                return back by the API.
        """
        graph_string = graph_dict.get('graph_elements')

        if isinstance(graph_string, str):
            try:
                self.resource_data = json.loads(graph_string)
            except json.JSONDecodeError as exc:
                raise ValueError('Unable to read graph data') from exc
        elif isinstance(graph_string, dict):
            self.resource_data = graph_string
        else:
            raise ValueError('Graph elements missing or not of correct value.')

        graph_config = graph_dict.get('graph_config')
        if graph_config:
            self._graph_config = json.loads(graph_config)

        self._created_at = dateutil.parser.parse(
            graph_dict.get('created_at', ''))
        self._updated_at = dateutil.parser.parse(
            graph_dict.get('updated_at', ''))

    def _serialize(self, config_dict):
        """Serialize a graph config dictionary as JSON.

        Due to networkx's use of numpy.ndarray to describe
        coordinates JSON is unable to serialize the dictionary.
        It is therefore needed to convert the ndarray into lists
        so that they can be serialized by JSON.

        Args:
            config_dict (dict): the graph config dictionary.

        Returns:
            A JSON serialized string with the dictionary's content.
        """
        for key, item in config_dict.get('layout', {}).items():
            if not isinstance(item, numpy.ndarray):
                continue
            config_dict['layout'][key] = list(item)

        return json.dumps(config_dict)

    @property
    def created_at(self):
        """Property that returns back the creation time of a graph."""
        if self._created_at:
            return self._created_at.isoformat()
        return self._created_at

    def delete(self):
        """Deletes the saved graph from the store."""
        if not self._resource_id:
            logger.warning(
                'Unable to delete the saved graph, it does not appear to be '
                'saved in the first place.')
            return False

        resource_url = (
            f'{self.api.api_root}/sketches/{self._sketch.id}/graphs/'
            f'{self._resource_id}/')

        response = self.api.session.delete(resource_url)
        return error.check_return_status(response, logger)

    @property
    def description(self):
        """Property that returns back the description of the saved search."""
        return self._description

    @description.setter
    def description(self, description):
        """Make changes to the saved search description field."""
        self._description = description
        self.commit()

    @property
    def draw(self):
        """Property that returns back a drawing with the graph."""
        if not self.graph:
            return None

        if not self.graph.size():
            return None

        data = self.data.get('elements', {})
        nodes = data.get('nodes', [])

        label_dict = {
            x.get('data', {}).get('name', ''): x.get(
                'data', {}).get('label', '') for x in nodes
        }

        edges = data.get('edges', [])
        edge_dict = {
            x.get('data', {}).get('name', ''): x.get(
                'data', {}).get('label', '') for x in edges
        }
        label_dict.update(edge_dict)

        label_keys = list(label_dict.keys())
        for key in label_keys:
            if not key:
                _ = label_dict.pop(key)

        return nx.draw_networkx(
            self.graph, with_labels=True, labels=label_dict,
            pos=self.layout)

    @property
    def graph(self):
        """Property that returns back a graph object."""
        if self._graph:
            return self._graph

        if not self.resource_data:
            return nx.Graph()

        # It is necessary to do a deep copy, otherwise the upstream
        # nx library modifies the resource data.
        graph_dict = copy.deepcopy(self.resource_data)
        self._graph = nx.cytoscape_graph(graph_dict)
        return self._graph

    @property
    def graph_config(self):
        """Property that returns the graph config."""
        if self._graph_config:
            return self._graph_config

        self._graph_config = {
            'filter': {
                'indices': [
                    t.index_name for t in self._sketch.list_timelines()],
            },
            'layout': self.layout
        }
        return self._graph_config

    @graph_config.setter
    def graph_config(self, graph_config):
        """Change the graph config."""
        if not isinstance(graph_config, dict):
            raise ValueError('Graph config needs to be a dict.')

        self._graph_config = graph_config

    def from_graph(self, graph_obj):
        """Initialize from a networkx graph object.

        Args:
            graph_obj (nx.Graph): a graph object.
        """
        if not graph_obj.size():
            logger.warning('Unable to load graph from an empty graph object.')
            return
        self.resource_data = nx.cytoscape_data(graph_obj)
        self._graph = graph_obj
        self._name = ''
        self._description = 'From a graph object.'
        time = datetime.datetime.now(datetime.timezone.utc)
        self._created_at = time
        self._updated_at = time

    def from_manual(self, data, **kwargs):  # pylint: disable=arguments-differ
        """Generate a new graph using a dictionary.

        Args:
            data (dict): A dictionary of dictionaries adjacency representation.
            kwargs (dict[str, object]): Depending on the resource they may
                require different sets of arguments to be able to run a raw
                API request.

        Raises:
            ValueError: If the import is not successful.
        """
        super().from_manual(**kwargs)
        if not isinstance(data, dict):
            raise ValueError('Data needs to be a dict of dictionaries')

        try:
            graph = nx.from_dict_of_dicts(data)
        except AttributeError as exc:
            raise ValueError('Unable to generate a graph') from exc

        self.from_graph(graph)

    def from_plugin(self, plugin_name, plugin_config=None, refresh=False):
        """Initialize the graph from a cached plugin graph.

        Args:
            plugin_name (str): the name of the graph plugin to use.
            plugin_config (dict): optional dictionary to configure the plugin.
            refresh (bool): optional bool that if set refreshes the graph,
                otherwise the graph is pulled from the cache, if it exists.
                Defaults to False, meaning it pulls from the cache if it
                exists.

        Raises:
            ValueError: If the plugin doesn't exist or some issues came up
                during processing.
        """
        plugin_names = [x.get('name', '') for x in self.plugins]
        if plugin_name.lower() not in plugin_names:
            raise ValueError(
                f'Plugin [{plugin_name}] not part of the supported plugins.')

        if plugin_config and not isinstance(plugin_config, dict):
            raise ValueError('Plugin config needs to be a dict.')

        resource_url = f'{self.api.api_root}/sketches/{self._sketch.id}/graph/'
        data = {
            'plugin': plugin_name,
            'config': plugin_config,
            'refresh': bool(refresh),
        }

        if self.timelines:
            data['timeline_ids'] = self.timelines

        if plugin_config:
            if isinstance(plugin_config, str):
                self._graph_config = json.loads(plugin_config)
            else:
                self._graph_config = plugin_config

        response = self.api.session.post(resource_url, json=data)
        status = error.check_return_status(response, logger)
        if not status:
            error.error_message(
                response, 'Unable to retrieve cached graph', error=RuntimeError)

        response_json = error.get_response_json(response, logger)
        cache_dict = response_json.get('objects', [{}])[0]
        self._parse_graph_dict(cache_dict)
        self._description = f'Graph created from the {plugin_name} plugin.'

    def from_saved(self, graph_id):  # pylint: disable=arguments-differ
        """Initialize the graph object from a saved graph.

        Args:
            graph_id: integer value for the saved graph (primary key).

        Raises:
            ValueError: If issues came up during processing.
        """
        resource_id = f'sketches/{self._sketch.id}/graphs/{graph_id}/'
        data = self.api.fetch_resource_data(resource_id)

        objects = data.get('objects')
        if not objects:
            logger.warning(
                'Unable to load saved graph with ID: %d, '
                'are you sure it exists?', graph_id)
        graph_dict = objects[0]
        self._parse_graph_dict(graph_dict)

        self._resource_id = graph_id
        self._name = graph_dict.get('name', 'No name')
        self._description = graph_dict.get('description', '')
        self._username = graph_dict.get('user', {}).get('username', 'System')
        self._created_at = dateutil.parser.parse(
            graph_dict.get('created_at', ''))
        self._updated_at = dateutil.parser.parse(
            graph_dict.get('updated_at', ''))

    @property
    def layout(self):
        """Property that returns back the layout of the graph."""
        if self._layout:
            return self._layout

        layout = self._GRAPH_LAYOUTS.get('spring')
        self._layout = layout(self.graph)
        return self._layout

    @layout.setter
    def layout(self, layout):
        """Change the layout manually."""
        if not isinstance(layout, dict):
            raise ValueError('Layout needs to be a dict.')
        self._layout = layout

    @property
    def layout_strings(self):
        """Property that returns a list of potential layouts to use."""
        return list(self._GRAPH_LAYOUTS.keys())

    @property
    def name(self):
        """Property that returns the query name."""
        return self._name

    @name.setter
    def name(self, name):
        """Make changes to the saved search name."""
        self._name = name
        self.commit()

    @property
    def plugins(self):
        """Property that returns back the supported plugins."""
        plugin_dict = self.api.fetch_resource_data('graphs/')
        return plugin_dict

    @property
    def plugins_table(self):
        """Property that returns a dataframe with the available plugins."""
        return pandas.DataFrame(self.plugins)

    def save(self):
        """Save the search in the database.

        Raises:
            ValueError: if there are values missing in order to save the query.
        """
        if not self.name:
            raise ValueError(
                'No name for the graph. Please select a name first.')

        if not self.description:
            logger.warning(
                'No description selected for graph, saving without one')

        if self._resource_id:
            resource_url = (
                f'{self.api.api_root}/sketches/{self._sketch.id}/graphs/'
                f'{self._resource_id}/')
        else:
            resource_url = (
                f'{self.api.api_root}/sketches/{self._sketch.id}/graphs/')

        cytoscape_json = nx.readwrite.json_graph.cytoscape_data(self.graph)
        elements = cytoscape_json.get('elements', [])
        element_list = []
        for group in elements:
            for element in elements[group]:
                element['group'] = group
                element_list.append(element)

        data = {
            'name': self.name,
            'graph_config': self._serialize(self.graph_config),
            'description': self.description,
            'elements': element_list,
        }

        response = self.api.session.post(resource_url, json=data)
        status = error.check_return_status(response, logger)
        if not status:
            error.error_message(
                response, 'Unable to save search', error=RuntimeError)

        response_json = error.get_response_json(response, logger)

        graph_dict = response_json.get('objects', [{}])[0]
        self._resource_id = graph_dict.get('id', 0)
        return f'Saved graph to ID: {self._resource_id}'

    def set_layout_type(self, layout_string):
        """Use a layout from the layout strings."""
        layout = self._GRAPH_LAYOUTS.get(layout_string)
        if layout:
            self.layout = layout(self.graph)

    @property
    def timelines(self):
        """Property that returns the set of timelines this graph uses."""
        return self._timelines

    @timelines.setter
    def timelines(self, timelines):
        """Sets the timelines."""
        if not isinstance(timelines, (list, tuple)):
            logger.error('Unable to add timelines, this needs to be a list')

        for timeline in timelines:
            if isinstance(timeline, int):
                self._timelines.append(timeline)
                continue

            if hasattr(timeline, 'id') and hasattr(timeline, 'index_name'):
                self._timelines.append(timeline.id)

    def to_dict(self):
        """Returns a dict with the graph content."""
        return self.resource_data

    def to_pandas(self):
        """Returns a pandas dataframe with the graph content."""
        if not self.resource_data:
            return pandas.DataFrame()

        data = self.resource_data
        elements = data.get('elements', {})
        nodes = elements.get('nodes', [])
        edges = elements.get('edges', [])

        node_list = [x.get('data') for x in nodes]
        edge_list = [x.get('data') for x in edges]
        node_df = pandas.DataFrame(node_list)
        node_df['type'] = 'node'
        edge_df = pandas.DataFrame(edge_list)
        edge_df['type'] = 'edge'
        return pandas.concat([node_df, edge_df])

    @property
    def updated_at(self):
        """Property that returns back the last updated time of a graph."""
        if self._updated_at:
            return self._updated_at.isoformat()
        return self._updated_at



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLI assistance for importer tools."""

from typing import Any
from typing import Callable
from typing import Optional
from typing import Text

import getpass


def ask_question(
        question: Text, input_type: Callable[[Text], Any],
        default: Optional[Any] = None,
        hide_input: Optional[bool] = False) -> Any:
    """Presents the user with a prompt with a default return value and a type.

    Args:
        question (str): the text that the user will be prompted.
        input_type (type): the type of the input data.
        default (object): default value for the question, optional.
        hide_input (bool): whether the input should be hidden, eg. when asking
            for a password.

    Returns:
        object: The value (type of input_type) that is ready by the user.
    """
    if hide_input:
        if default:
            hint = '**'
        else:
            hint = ''

        return getpass.getpass(f'{question} [{hint}] ')

    if default:
        ask = f'{question} [{default}]'
    else:
        ask = question

    answer = input(f'{ask}: ')
    return input_type(answer)


def confirm_choice(
        choice: Text, default: Optional[bool] = True,
        abort: Optional[bool] = True) -> bool:
    """Returns a bool from a yes/no question presented to the end user.

    Args:
        choice (str): the question presented to the end user.
        default (bool): the default for the confirmation answer. If True the
            default is Y(es), if False the default is N(o)
        abort (bool): if the program should abort if the user answer to the
            confirm prompt is no. The default is an abort.

    Raises:
        RuntimeError: If abort is set to True and the choice is no.

    Returns:
        bool: False if the user entered no, True if the user entered yes
    """
    if default:
        hint = 'Y/n'
    else:
        hint = 'y/N'
    answer = input(f'{choice} [{hint}]: ')

    value = None
    if answer.lower() in ['y', 'yes']:
        value = True

    if answer.lower() in ['n', 'no']:
        value = False

    if not answer:
        value = default

    if value is None:
        print('Invalid answer')
        return confirm_choice(choice=choice, default=default, abort=abort)

    if abort and not value:
        raise RuntimeError('Aborting')

    return value



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API client library."""
import json


class BaseResource:
    """Base resource object."""

    def __init__(self, api, resource_uri):
        """Initialize object.

        Args:
            api: An instance of TimesketchApi object.
            resource_uri: The URI part of the API resource to call.
        """
        self.api = api
        self.resource_uri = resource_uri
        self.resource_data = None

    def lazyload_data(self, refresh_cache=False):
        """Load resource data once and cache the result.

        Args:
            refresh_cache: Boolean indicating if to update cache.

        Returns:
            Dictionary with resource data.
        """
        if not self.resource_data or refresh_cache:
            self.resource_data = self.api.fetch_resource_data(self.resource_uri)
        return self.resource_data

    @property
    def data(self):
        """Property to fetch resource data.

        Returns:
            Dictionary with resource data.
        """
        return self.lazyload_data()


class SketchResource(BaseResource):
    """Sketch resource object."""

    def __init__(self, resource_uri, sketch):
        """Initialize the sketch resource object.

        Args:
            resource_uri: The URI part of the API resource to call.
            sketch: The sketch object of the sketch this resource is tied to.
        """
        super().__init__(sketch.api, resource_uri)

        self._labels = []
        self._resource_id = 0
        self._sketch = sketch
        self._username = ''

    def _get_top_level_attribute(self, name, default_value=None, refresh=False):
        """Returns a top level attribute from a resource object.

        Args:
            name: String with the attribute name.
            default_value: The default value if the attribute does not exit,
                defaults to None.
            refresh: If set to True then the data will be refreshed.

        Returns:
            The dict value of the key "name".
        """
        resource = self.lazyload_data(refresh_cache=refresh)
        resource_objects = resource.get('objects')
        if not resource_objects:
            return default_value

        if not len(resource_objects) == 1:
            return default_value

        first_object = resource_objects[0]
        return first_object.get(name, default_value)

    def add_label(self, label):
        """Add a label to the resource.

        Args:
            label (str): string with the label information.
        """
        if label in self._labels:
            return
        self._labels.append(label)
        self.save()

    def commit(self):
        """Calls the save function if the object has already been saved."""
        if not self._resource_id:
            return
        self.save()

    def delete(self):
        """Deletes the resource from the list of stored resources."""
        raise NotImplementedError

    @property
    def dict(self):
        """Property that returns back a Dict with the results."""
        return self.to_dict()

    def from_manual(self, **kwargs):
        """Initialize the resource object by running a raw API request.

        The API request functionality should be implemented by other functions
        that inherit this as a base class.

        Args:
            kwargs (dict[str, object]): Depending on the resource they may
                require different sets of arguments to be able to run a raw
                API request.

        Raises:
            ValueError: If there are any unused keyword arguments passed to
                the function.
        """
        if kwargs:
            raise ValueError('Unused keyword arguments: {0:s}.'.format(
                ', '.join(kwargs.keys())))

    def from_saved(self, resource_id):
        """Initialize the resource object from a saved resource.

        Args:
            resource_id: integer value for the saved resource (primary key).
        """
        raise NotImplementedError

    @property
    def id(self):
        """Property that returns back the resource ID."""
        return self._resource_id

    @property
    def json(self):
        """Property that returns back a JSON object with the results."""
        return json.dumps(self.dict)

    @property
    def labels(self):
        """Property that returns a list of the resource labels."""
        return self._labels

    @property
    def sketch(self):
        """Property that returns the sketch object."""
        return self._sketch

    @property
    def table(self):
        """Property that returns a pandas DataFrame."""
        return self.to_pandas()

    @property
    def user(self):
        """Property that returns the username of who ran the aggregation."""
        if not self._username:
            return 'System'
        return self._username

    def to_dict(self):
        """Returns a dict."""
        raise NotImplementedError

    def to_pandas(self):
        """Returns a pandas DataFrame."""
        raise NotImplementedError

    def save(self):
        """Sends a request to save the resource."""
        raise NotImplementedError



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Timesketch API client"""
from __future__ import unicode_literals

import unittest
import mock

from . import test_lib
from . import client

MOCK_SIGMA_RULE = """
title: Suspicious Installation of Zenmap
id: 5266a592-b793-11ea-b3de-0242ac130004
description: Detects suspicious installation of Zenmap
references:
    - https://rmusser.net/docs/ATT&CK-Stuff/ATT&CK/Discovery.html
author: Alexander Jaeger
date: 2020/06/26
modified: 2021/01/01
logsource:
    product: linux
    service: shell
detection:
    keywords:
        # Generic suspicious commands
        - '*apt-get install zmap*'
    condition: keywords
falsepositives:
    - Unknown
level: high
"""

MOCK_SIGMA_RULE_ERROR1 = """
title: Suspicious Foobar
id: 5266a592-b793-11ea-b3de-0242ac130004
description: Detects suspicious installation of Zenmap
references:
    - https://rmusser.net/docs/ATT&CK-Stuff/ATT&CK/Discovery.html
author: Alexander Jaeger
date: 2020/06/26
modified: 2020/06/26
"""

class TimesketchSigmaTest(unittest.TestCase):
    """Test Sigma."""

    @mock.patch('requests.Session', test_lib.mock_session)
    def setUp(self):
        """Setup test case."""
        self.api_client = client.TimesketchApi(
            'http://127.0.0.1', 'test', 'test')

    def test_sigma_rule(self):
        """Test single Sigma rule."""

        rule = self.api_client.get_sigma_rule(
            rule_uuid='5266a592-b793-11ea-b3de-0242ac130004')
        rule.from_rule_uuid('5266a592-b793-11ea-b3de-0242ac130004')
        self.assertGreater(len(rule.attributes),5)
        self.assertIsNotNone(rule)
        self.assertIn('Alexander', rule.author)
        self.assertIn('Alexander', rule.get_attribute('author'))
        self.assertEqual(rule.id, '5266a592-b793-11ea-b3de-0242ac130004')
        self.assertEqual(rule.title, 'Suspicious Installation of Zenmap')
        self.assertIn('zmap', rule.es_query, 'ES_Query does not match')
        self.assertIn('b793', rule.id)
        self.assertIn('/syslog/foobar/', rule.file_relpath)
        self.assertIn('sigma/rule/5266a592', rule.resource_uri)
        self.assertIn('suspicious installation of Zenmap', rule.description)
        self.assertIn('high', rule.level)
        self.assertEqual(len(rule.falsepositives), 1)
        self.assertIn('Unknown', rule.falsepositives[0])
        self.assertIn('susp_zenmap', rule.file_name)
        self.assertIn('2020/06/26', rule.date)
        self.assertIn('2021/01/01', rule.modified)
        self.assertIn('high', rule.level)
        self.assertIn('foobar.com', rule.references[0])
        self.assertEqual(len(rule.detection), 2)
        self.assertEqual(len(rule.logsource), 2)


    def test_sigma_rules(self):
        '''Testing the Sigma rules list'''

        rules = self.api_client.list_sigma_rules()
        self.assertIsNotNone(rules)
        self.assertEqual(len(rules), 2)
        rule = rules[0]
        self.assertEqual(rule.title, 'Suspicious Installation of Zenmap')
        self.assertIn('zmap', rule.es_query, 'ES_Query does not match')
        self.assertIn('b793', rule.id)
        self.assertIn('Alexander', rule.author)
        self.assertIn('2020/06/26',rule.date)
        self.assertIn('installation of Zenmap', rule.description)
        self.assertEqual(len(rule.detection), 2)
        self.assertIn('zmap*', rule.es_query)
        self.assertIn('Unknown', rule.falsepositives[0])
        self.assertEqual(len(rule.detection), 2)
        self.assertEqual(len(rule.logsource), 2)
        self.assertIn('2020/06/26', rule.modified)
        self.assertIn('/linux/syslog/foobar', rule.file_relpath)
        self.assertIn('lnx_susp_zenmap', rule.file_name)
        self.assertIn('high', rule.level)
        self.assertIn('foobar.com', rule.references[0])


    def test_get_sigma_rule_by_text(self):

        rule = self.api_client.get_sigma_rule_by_text(MOCK_SIGMA_RULE)

        self.assertIsNotNone(rule)
        self.assertGreater(len(rule.attributes),5)
        self.assertIn('zsh', rule.es_query)
        self.assertIn('Installation of foobar', rule.title)
        self.assertIn('', rule.id)
        self.assertIn('', rule.file_relpath)
        self.assertIn('http://127.0.0.1/api/v1/sigma/text/', rule.resource_uri)
        self.assertIn('suspicious installation of foobar', rule.description)
        self.assertIn('high', rule.level)
        self.assertEqual(len(rule.falsepositives), 1)
        self.assertIn('Unknown', rule.falsepositives[0])
        self.assertIn('N/A', rule.file_name)
        self.assertIn('Alexander', rule.author)
        self.assertIn('2020/12/10', rule.date)
        self.assertIn('2021/01/01', rule.modified)
        self.assertEqual(len(rule.detection), 2)
        self.assertEqual(len(rule.logsource), 2)



# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Timesketch API client"""
from __future__ import unicode_literals

import json


def mock_session():
    """Mock HTTP requests session."""

    class MockHeaders:
        """Mock requests HTTP headers."""

        # pylint: disable=unused-argument
        @staticmethod
        def update(*args, **kwargs):
            """Mock header update method."""
            return

    class MockSession:
        """Mock HTTP requests session."""

        def __init__(self):
            """Initializes the mock Session object."""
            self.verify = False
            self.headers = MockHeaders()
            self._post_done = False

        # pylint: disable=unused-argument
        @staticmethod
        def get(*args, **kwargs):
            """Mock GET request handler."""
            return mock_response(*args, **kwargs)

        # pylint: disable=unused-argument
        def post(self, *args, **kwargs):
            """Mock POST request handler."""
            if self._post_done:
                return mock_response(*args, empty=True)
            return mock_response(*args, **kwargs)

    return MockSession()


# pylint: disable=unused-argument
def mock_response(*args, **kwargs):
    """Mocks HTTP response."""

    class MockResponse:
        """Mock HTTP response object."""

        def __init__(self, json_data=None, text_data=None, status_code=200):
            """Initializes mock object."""
            self.json_data = json_data
            self.text = text_data
            self.status_code = status_code

        def json(self):
            """Mock JSON response."""
            return self.json_data

    auth_text_data = '<input id="csrf_token" name="csrf_token" value="test">'

    archive_data = {
        'is_archived': False,
        'sketch_id': 1,
        'sketch_name': 'test',
    }

    sketch_data = {
        'meta': {
            'views': [{
                'id': 1,
                'name': 'test'
            }, {
                'id': 2,
                'name': 'test'
            }],
            'es_time': 41444,
        },
        'objects': [{
            'id':
            1,
            'name': 'test',
            'description': 'test',
            'timelines': [{
                'id': 1,
                'name': 'test',
                'searchindex': {
                    'index_name': 'test'
                }
            }, {
                'id': 2,
                'name': 'test',
                'searchindex': {
                    'index_name': 'test'
                }
            }]
        }]
    }

    view_data_1 = {
        'objects': [{
            'id': 1,
            'name': 'test',
            'description': 'meant for testing purposes only.',
            'user': {'username': 'gisli'},
            'query_string': 'test:"foobar"',
            'query_dsl': '',
            'searchtemplate': '',
            'aggregation': '',
            'created_at': '2020-11-30T15:17:29',
            'updated_at': '2020-11-30T15:17:29',
        }],
    }

    view_data_2 = {
        'objects': [{
            'id': 2,
            'name': 'more test',
            'description': 'really meant for testing purposes only.',
            'user': {'username': 'eirikur'},
            'query_string': 'test:"bar"',
            'query_dsl': '',
            'searchtemplate': '',
            'aggregation': '',
            'created_at': '2020-11-30T15:17:29',
            'updated_at': '2020-11-30T15:17:29',
        }],
    }

    sketch_list_data = {
        'meta': {
            'es_time': 324,
            'total_pages': 1,
            'current_page': 1
        },
        'objects': sketch_data['objects']}

    timeline_data = {
        'meta': {
            'es_time': 12,
        },
        'objects': [{
            'id': 1,
            'name': 'test',
            'searchindex': {
                'id': 1234,
                'index_name': 'test'
            }
        }]
    }

    more_timeline_data = {
        'meta': {
            'es_time': 12,
        },
        'objects': [{
            'id': 2,
            'name': 'more_test',
            'searchindex': {
                'id': 42,
                'index_name': 'even_more_test'
            }
        }]
    }

    empty_data = {
        'meta': {'es_time': 0},
        'objects': []
    }

    story_list_data = {
        'meta': {'es_time': 23},
        'objects': [[{'id': 1}]]
    }

    story_data = {
        'meta': {
            'es_time': 1,
        },
        'objects': [{
            'title': 'My First Story',
            'content': json.dumps([
                {
                    'componentName': '',
                    'componentProps': {},
                    'content': '# My Heading\nWith Some Text.',
                    'edit': False,
                    'showPanel': False,
                    'isActive': False
                },
                {
                    'componentName': 'TsViewEventList',
                    'componentProps': {
                        'view': {
                            'id': 1,
                            'name': 'Smoking Gun'}},
                    'content': '',
                    'edit': False,
                    'showPanel': False,
                    'isActive': False
                },
                {
                    'componentName': '',
                    'componentProps': {},
                    'content': '... and that was the true crime.',
                    'edit': False,
                    'showPanel': False,
                    'isActive': False
                }
            ])
        }]
    }

    sigma_list = {
        'meta': {
            'current_user': 'dev',
            'rules_count': 2
        },
        'objects': [
            {
                'author': 'Alexander Jaeger',
                'date': '2020/06/26',
                'description': 'Detects suspicious installation of Zenmap',
                'detection': {
                    'condition': 'keywords',
                    'keywords': ['*apt-get install zmap*']
                    },
                'falsepositives': ['Unknown'],
                'id': '5266a592-b793-11ea-b3de-0242ac130004',
                'level': 'high',
                'logsource': {
                    'product': 'linux', 'service': 'shell'
                    },
                'es_query': '("*apt\\-get\\ install\\ zmap*")',
                'modified': '2020/06/26',
                'references': ['httpx://foobar.com'],
                'title': 'Suspicious Installation of Zenmap',
                'file_name': 'lnx_susp_zenmap',
                'file_relpath' : '/linux/syslog/foobar/'
            }, {
                'author': 'Alexander Jaeger',
                'date': '2020/11/10',
                'description': 'Detects suspicious installation of foobar',
                'detection': {
                    'condition': 'keywords',
                    'keywords': ['*apt-get install foobar*']
                    },
                'falsepositives': ['Unknown'],
                'id': '776bdd11-f3aa-436e-9d03-9d6159e9814e',
                'level': 'high',
                'logsource': {
                    'product': 'linux', 'service': 'shell'
                    },
                'es_query': '("*apt\\-get\\ install\\ foo*")',
                'modified': '2020/06/26',
                'references': ['httpx://foobar.com'],
                'title': 'Suspicious Installation of Zenmap',
                'file_name': 'lnx_susp_foobar',
                'file_relpath' : '/windows/foobar/'
                }
        ]
    }

    sigma_rule = {
        'meta': {
            'parsed': True
        },
        'objects':[
            {
            'title': 'Suspicious Installation of Zenmap',
            'id': '5266a592-b793-11ea-b3de-0242ac130004',
            'description': 'Detects suspicious installation of Zenmap',
            'references': ['httpx://foobar.com'],
            'author': 'Alexander Jaeger',
            'date': '2020/06/26',
            'modified': '2021/01/01',
            'logsource': {
                'product': 'linux', 'service': 'shell'
                },
            'detection': {
                'keywords': ['*apt-get install zmap*'],
                'condition': 'keywords'
                },
            'falsepositives': ['Unknown'],
            'level': 'high',
            'es_query': '("*apt\\-get\\ install\\ zmap*")',
            'file_name': 'lnx_susp_zenmap',
            'file_relpath' : '/linux/syslog/foobar/'
            }
        ]
    }
    sigma_rule_text_mock = {
        'meta': {
            'parsed': True
        },
        'objects':[
            {
                'title': 'Installation of foobar',
                'id': 'bb1e0d1d-cd13-4b65-bf7e-69b4e740266b',
                'description': 'Detects suspicious installation of foobar',
                'references': ['https://samle.com/foobar'],
                'author': 'Alexander Jaeger',
                'date': '2020/12/10',
                'modified': '2021/01/01',
                'logsource': {
                    'product': 'linux',
                    'service': 'shell'
                },
                'detection': {
                    'keywords': ['*apt-get install foobar*'],
                    'condition': 'keywords'
                },
                'falsepositives': ['Unknown'],
                'level': 'high',
                'es_query':
                    '(data_type:("shell\\:zsh\\:history" OR "bash\\:history\\:command" OR "apt\\:history\\:line" OR "selinux\\:line") AND "*apt\\-get\\ install\\ foobar*")',# pylint: disable=line-too-long
                'file_name': 'N/A',
                'file_relpath': 'N/A'
            }
        ]
    }

    # Register API endpoints to the correct mock response data.
    url_router = {
        'http://127.0.0.1':
        MockResponse(text_data=auth_text_data),
        'http://127.0.0.1/api/v1/sketches/':
        MockResponse(json_data=sketch_list_data),
        'http://127.0.0.1/api/v1/sketches/1':
        MockResponse(json_data=sketch_data),
        'http://127.0.0.1/api/v1/sketches/1/views/1/':
        MockResponse(json_data=view_data_1),
        'http://127.0.0.1/api/v1/sketches/1/views/2/':
        MockResponse(json_data=view_data_2),
        'http://127.0.0.1/api/v1/sketches/1/timelines/1/':
        MockResponse(json_data=timeline_data),
        'http://127.0.0.1/api/v1/sketches/1/timelines/2/':
        MockResponse(json_data=more_timeline_data),
        'http://127.0.0.1/api/v1/sketches/1/explore/':
        MockResponse(json_data=timeline_data),
        'http://127.0.0.1/api/v1/sketches/1/stories/':
        MockResponse(json_data=story_list_data),
        'http://127.0.0.1/api/v1/sketches/1/stories/1/':
        MockResponse(json_data=story_data),
        'http://127.0.0.1/api/v1/sketches/1/archive/':
        MockResponse(json_data=archive_data),
        'http://127.0.0.1/api/v1/sigma/rule/5266a592-b793-11ea-b3de-0242ac130004':# pylint: disable=line-too-long
        MockResponse(json_data=sigma_rule),
        'http://127.0.0.1/api/v1/sigma/':
        MockResponse(json_data=sigma_list),
        'http://127.0.0.1/api/v1/sigma/text/':
        MockResponse(json_data=sigma_rule_text_mock),
    }

    if kwargs.get('empty', False):
        return MockResponse(text_data=empty_data)

    return url_router.get(args[0], MockResponse(None, 404))



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API sigma library."""
from __future__ import unicode_literals

import logging

from . import resource
from . import error

logger = logging.getLogger('timesketch_api.sigma')


class Sigma(resource.BaseResource):
    """Timesketch sigma object.

    A sigma object in Timesketch is a collection of one or more rules.

    Attributes:
        rule_uuid: The ID of the rule.
    """


    def __init__(self, api):
        """Initializes the Sigma object.

        Args:
            api: An instance of TimesketchApi object.

        """
        self._attr_dict = {}
        resource_uri = 'sigma/'
        super().__init__(
            api=api, resource_uri=resource_uri)

    @property
    def attributes(self):
        """Returns a list of all attribute keys for the rule"""
        return list(self._attr_dict.keys())

    def get_attribute(self, key):
        """Get a value for a given key in case it has no dedicated property"""
        if not self._attr_dict:
            return ''
        return self._attr_dict.get(key, '')

    @property
    def es_query(self):
        """Returns the ElasticSearch query."""
        return self.get_attribute('es_query')

    @property
    def title(self):
        """Returns the sigma rule title."""
        return self.get_attribute('title')

    @property
    def id(self):
        """Returns the sigma rule id."""
        return self.get_attribute('id')

    @property
    def file_relpath(self):
        """Returns the relative filepath of the rule."""
        return self.get_attribute('file_relpath')

    @property
    def rule_uuid(self):
        """Returns the rule id."""
        return self.get_attribute('rule_uuid')

    @property
    def file_name(self):
        """Returns the rule filename."""
        return self.get_attribute('file_name')

    @property
    def description(self):
        """Returns the rule description."""
        return self.get_attribute('description')

    @property
    def level(self):
        """Returns the rule confidence level."""
        return self.get_attribute('level')

    @property
    def falsepositives(self):
        """Returns the rule falsepositives."""
        return self.get_attribute('falsepositives')

    @property
    def author(self):
        """Returns the rule author."""
        return self.get_attribute('author')

    @property
    def date(self):
        """Returns the rule date."""
        return self.get_attribute('date')

    @property
    def modified(self):
        """Returns the rule modified date."""
        return self.get_attribute('modified')


    @property
    def logsource(self):
        """Returns the rule logsource."""
        return self.get_attribute('logsource')

    @property
    def detection(self):
        """Returns the rule detection."""
        return self.get_attribute('detection')

    @property
    def references(self):
        """Returns the rule references."""
        return self.get_attribute('references')

    def set_value(self, key, value):
        """Sets the value for a given key

        Args:
            key: key to set the value
            value: value to set

        """
        self._attr_dict[key] = value

    def _load_rule_dict(self, rule_dict):
        """Load a dict into a rule"""
        for key, value in rule_dict.items():
            self.set_value(key, value)

    def from_rule_uuid(self, rule_uuid):
        """Get a Sigma object from a rule uuid.

        Args:
            rule_uuid: Id of the sigma rule.

        """
        self.resource_uri = f'sigma/rule/{rule_uuid}'
        super().__init__(
            api=self.api, resource_uri=self.resource_uri)

        self.lazyload_data(refresh_cache=True)
        objects = self.data.get('objects')
        if not objects:
            logger.error('Unable to parse rule with given text')
            raise ValueError('No rules found.')
        rule_dict = objects[0]
        for key, value in rule_dict.items():
            self.set_value(key, value)

    def from_text(self, rule_text):
        """Get a Sigma object from a rule text.

        Args:
            rule_text: Rule text to be parsed.

        Raises:
            ValueError: If no response was given
        """
        self.resource_uri = '{0:s}/sigma/text/'.format(self.api.api_root)
        data = {'title': 'Get_Sigma_by_text', 'content': rule_text}
        response = self.api.session.post(self.resource_uri, json=data)
        response_dict = error.get_response_json(response, logger)

        objects = response_dict.get('objects')
        if not objects:
            logger.warning(
                'Unable to parse rule with given text')
            raise ValueError('No rules found.')

        rule_dict = objects[0]
        for key, value in rule_dict.items():
            self.set_value(key, value)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API search object."""
import datetime
import json
import logging

import pandas

from . import error
from . import resource
from . import searchtemplate


logger = logging.getLogger('timesketch_api.search')


class Chip:
    """Class definition for a query filter chip."""

    # The type of a chip that is defined.
    CHIP_TYPE = ''

    # The chip value defines what property or attribute of the
    # chip class will be used to generate the chip value.
    CHIP_VALUE = ''

    # The value of the chip field.
    CHIP_FIELD = ''

    def __init__(self):
        """Initialize the chip."""
        self._active = True
        self._operator = 'must'
        self._chip_field = self.CHIP_FIELD

    @property
    def active(self):
        """A property that returns whether the chip is active or not."""
        return self._active

    @active.setter
    def active(self, active):
        """Decide whether the chip is active or disabled."""
        self._active = bool(active)

    @property
    def chip(self):
        """A property that returns the chip value."""
        return {
            'field': self._chip_field,
            'type': self.CHIP_TYPE,
            'operator': self._operator,
            'active': self._active,
            'value': getattr(self, self.CHIP_VALUE, ''),
        }

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        raise NotImplementedError

    def set_include(self):
        """Configure the chip so the content needs to be included in results."""
        self._operator = 'must'

    def set_exclude(self):
        """Configure the chip so content needs to be excluded in results."""
        self._operator = 'must_not'

    def set_optional(self):
        """Configure the chip so the content is optional in results."""
        self._operator = 'should'

    def set_active(self):
        """Set the chip as active."""
        self._active = True

    def set_disable(self):
        """Disable the chip."""
        self._active = False


class DateIntervalChip(Chip):
    """A date interval chip."""

    CHIP_TYPE = 'datetime_interval'
    CHIP_VALUE = 'interval'

    _DATE_FORMAT = '%Y-%m-%dT%H:%M:%S'

    def __init__(self):
        """Initialize the chip."""
        super().__init__()
        self._date = None
        self._before = 5
        self._after = 5
        self._unit = 'm'

    def add_interval(self, before, after=None, unit='m'):
        """Set the interval of the chip.

        Args:
            before (int): the number of units that should be included
                before the date.
            after (int): optional number of units after the date. If not
                provided the value of before is used.
            unit (str): optional string with the unit of interval. This can
                be s for seconds, m for minutes, d for days and h for hours.
                The default value is m (minutes).

        Raises:
            ValueError if the unit is not correctly formed.
        """
        if after is None:
            after = before

        self.unit = unit

        self._before = before
        self._after = after

    @property
    def after(self):
        """Property that returns the time interval after the date."""
        return self._after

    @after.setter
    def after(self, after):
        """Make changes to the time interval after the date."""
        self._after = after

    @property
    def before(self):
        """Property that returns the time interval before the date."""
        return self._before

    @before.setter
    def before(self, before):
        """Make changes to the time interval before the date."""
        self._before = before

    @property
    def date(self):
        """Property that returns back the date."""
        if not self._date:
            return ''
        return self._date.strftime(self._DATE_FORMAT)

    @date.setter
    def date(self, date):
        """Make changes to the date."""
        try:
            dt = datetime.datetime.strptime(date, self._DATE_FORMAT)
        except ValueError as exc:
            logger.error(
                'Unable to add date chip, wrong date format', exc_info=True)
            raise ValueError('Wrong date format') from exc
        self._date = dt

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        value = chip_dict.get('value')
        if not value:
            return
        date, before, after = value.split()
        self.unit = before[-1]
        self.date = date
        self.before = int(before[1:-1])
        self.after = int(after[1:-1])

    @property
    def interval(self):
        """A property that returns back the full interval."""
        return (
            f'{self.date} -{self.before}{self.unit} +{self.after}{self.unit}')

    @property
    def unit(self):
        """Property that returns back the unit used."""
        return self._unit

    @unit.setter
    def unit(self, unit):
        """Make changes to the unit."""
        if unit not in ('s', 'm', 'd', 'h'):
            raise ValueError(
                'Unable to add interval, needs to be one of: '
                's (seconds), m (minutes), h (hours) or d (days)')
        self._unit = unit


class DateRangeChip(Chip):
    """A date range chip."""

    CHIP_TYPE = 'datetime_range'
    CHIP_VALUE = 'date_range'

    _DATE_FORMAT = '%Y-%m-%dT%H:%M:%S'

    def __init__(self):
        """Initialize the date range."""
        super().__init__()
        self._start_date = None
        self._end_date = None

    def add_end_time(self, end_time):
        """Add an end time to the range.

        Args:
            end_time (str): date string using the format '%Y-%m-%dT%H:%M:%s'

        Raises:
            ValueError: if the date format is incorrectly formatted.
        """
        try:
            dt = datetime.datetime.strptime(end_time, self._DATE_FORMAT)
        except ValueError as exc:
            logger.error(
                'Unable to add date chip, wrong date format', exc_info=True)
            raise ValueError('Wrong date format') from exc
        self._end_date = dt

    def add_start_time(self, start_time):
        """Add a start time to the range.

        Args:
            start_time (str): date string using the format '%Y-%m-%dT%H:%M:%s'

        Raises:
            ValueError: if the date format is incorrectly formatted.
        """
        try:
            dt = datetime.datetime.strptime(start_time, self._DATE_FORMAT)
        except ValueError as exc:
            logger.error(
                'Unable to add date chip, wrong date format', exc_info=True)
            raise ValueError('Wrong date format') from exc
        self._start_date = dt

    @property
    def end_time(self):
        """Property that returns the end time of a range."""
        if not self._end_date:
            return ''
        return self._end_date.strftime(self._DATE_FORMAT)

    @end_time.setter
    def end_time(self, end_time):
        """Sets the new end time."""
        self.add_end_time(end_time)

    @property
    def date_range(self):
        """Property that returns back the range."""
        return f'{self.start_time},{self.end_time}'

    @date_range.setter
    def date_range(self, date_range):
        """Sets the new range of the date range chip."""
        start_time, end_time = date_range.split(',')
        self.add_start_time(start_time)
        self.add_end_time(end_time)

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        chip_value = chip_dict.get('value')
        if not chip_value:
            return
        start, end = chip_value.split(',')
        self.start_time = start
        self.end_time = end

    @property
    def start_time(self):
        """Property that returns the start time of a range."""
        if not self._start_date:
            return ''
        return self._start_date.strftime(self._DATE_FORMAT)

    @start_time.setter
    def start_time(self, start_time):
        """Sets the new start time of a range."""
        self.add_start_time(start_time)


class LabelChip(Chip):
    """Label chip."""

    CHIP_TYPE = 'label'
    CHIP_VALUE = 'label'

    def __init__(self):
        """Initialize the chip."""
        super().__init__()
        self._label = ''

    def from_dict(self, chip_dict):
        """Configure the chip from a dictionary."""
        chip_value = chip_dict.get('value')
        if not chip_value:
            return

        self.label = chip_value

    @property
    def label(self):
        """Property that returns back the label."""
        return self._label

    @label.setter
    def label(self, label):
        """Make changes to the label."""
        self._label = label

    def use_comment_label(self):
        """Use the comment label."""
        self._label = '__ts_comment'

    def use_star_label(self):
        """Use the star label."""
        self._label = '__ts_star'


class TermChip(Chip):
    """Term chip definition."""

    CHIP_TYPE = 'term'
    CHIP_VALUE = 'query'

    def __init__(self):
        """Initialize the chip."""
        super().__init__()
        self._query = ''

    @property
    def field(self):
        """Property that returns back the field used to match against."""
        return self._chip_field

    @field.setter
    def field(self, field):
        """Make changes to the field used to match against."""
        self._chip_field = field

    def from_dict(self, chip_dict):
        """Configure the term chip from a dictionary."""
        chip_value = chip_dict.get('value')
        if not chip_value:
            return

        self.field = chip_dict.get('field')
        self.query = chip_value

    @property
    def query(self):
        """Property that returns back the query."""
        return self._query

    @query.setter
    def query(self, query):
        """Make changes to the query."""
        self._query = query


class Search(resource.SketchResource):
    """Search object."""

    DEFAULT_SIZE_LIMIT = 10000

    def __init__(self, sketch):
        resource_uri = f'sketches/{sketch.id}/explore/'
        super().__init__(sketch=sketch, resource_uri=resource_uri)

        self._aggregations = ''
        self._chips = []
        self._created_at = ''
        self._description = ''
        self._indices = '_all'
        self._max_entries = self.DEFAULT_SIZE_LIMIT
        self._name = ''
        self._query_dsl = ''
        self._query_filter = {}
        self._query_string = ''
        self._raw_response = None
        self._return_fields = ''
        self._scrolling = None
        self._searchtemplate = ''
        self._total_elastic_size = 0
        self._updated_at = ''

    def _extract_chips(self, query_filter):
        """Extract chips from a query_filter."""
        chips = query_filter.get('chips', [])
        if not chips:
            return

        for chip_dict in chips:
            chip_type = chip_dict.get('type')
            if not chip_type:
                continue

            chip = None
            if chip_type == 'datetime_interval':
                chip = DateIntervalChip()
            elif chip_type == 'datetime_range':
                chip = DateRangeChip()
            elif chip_type == 'label':
                chip = LabelChip()
            elif chip_type == 'term':
                chip = TermChip()

            if not chip:
                continue
            chip.from_dict(chip_dict)

            active = chip_dict.get('active', True)
            chip.active = active

            operator = chip_dict.get('operator', 'must')
            if operator == 'must':
                chip.set_include()
            elif operator == 'must_not':
                chip.set_exclude()

            self.add_chip(chip)

    def _execute_query(self, file_name='', count=False):
        """Execute a search request and store the results.

        Args:
            file_name (str): optional file path to a filename that
                all the results will be saved to. If not provided
                the results will be stored in the search object.
            count (bool): optional boolean that determines whether
                we want to execute the query or only count the
                number of events that the query would produce.
        """
        query_filter = self.query_filter
        if not isinstance(query_filter, dict):
            raise ValueError(
                'Unable to query with a query filter that isn\'t a dict.')

        stop_size = self._max_entries
        scrolling = not bool(stop_size and (
            stop_size < self.DEFAULT_SIZE_LIMIT))

        if self.scrolling is not None:
            scrolling = self.scrolling

        form_data = {
            'query': self._query_string,
            'filter': query_filter,
            'dsl': self._query_dsl,
            'count': count,
            'fields': self._return_fields,
            'enable_scroll': scrolling,
            'file_name': file_name,
        }

        response = self.api.session.post(
            f'{self.api.api_root}/{self.resource_uri}', json=form_data)
        if not error.check_return_status(response, logger):
            error.error_message(
                response, message='Unable to query results',
                error=ValueError)

        if file_name:
            with open(file_name, 'wb') as fw:
                fw.write(response.content)
            return

        response_json = error.get_response_json(response, logger)

        if count:
            meta = response_json.get('meta', {})
            self._total_elastic_size = meta.get('total_count', 0)
            return

        scroll_id = response_json.get('meta', {}).get('scroll_id', '')
        form_data['scroll_id'] = scroll_id

        count = len(response_json.get('objects', []))
        total_count = count
        while count > 0:
            if self._max_entries and total_count >= self._max_entries:
                break

            if not scroll_id:
                logger.debug('No scroll ID, will stop.')
                break

            more_response = self.api.session.post(
                f'{self.api.api_root}/{self.resource_uri}', json=form_data)
            if not error.check_return_status(more_response, logger):
                error.error_message(
                    response, message='Unable to query results',
                    error=ValueError)
            more_response_json = error.get_response_json(more_response, logger)
            count = len(more_response_json.get('objects', []))
            total_count += count
            response_json['objects'].extend(
                more_response_json.get('objects', []))
            more_meta = more_response_json.get('meta', {})
            added_time = more_meta.get('es_time', 0)
            response_json['meta']['es_time'] += added_time

        self._total_elastic_size = response_json.get(
            'meta', {}).get('es_total_count', 0)
        if self._total_elastic_size != total_count:
            logger.info(
                '%d results were returned, but '
                '%d records matched the search query',
                total_count, self._total_elastic_size)

        self._raw_response = response_json

    def add_chip(self, chip):
        """Add a chip to the ..."""
        self._chips.append(chip)
        self.commit()

    def add_date_range(self, start_time, end_time):
        """Add a date range chip to the search query.

        Args:
            start_time (str): a string with the start time of the range,
                the format should be '%Y-%m-%dT%H:%M:%S'
            end_time (str): a string with the end time of the range,
                the format should be '%Y-%m-%dT%H:%M:%S'
        """
        chip = DateRangeChip()
        chip.start_time = start_time
        chip.end_time = end_time
        self.add_chip(chip)

    @property
    def chips(self):
        """Property that returns all the chips in the search object."""
        return self._chips

    def commit(self):
        """Commit changes to the search object."""
        self._raw_response = None
        super().commit()

    @property
    def created_at(self):
        """Property that returns back the creation time of a search."""
        return self._created_at

    def delete(self):
        """Deletes the saved search from the store."""
        if not self._resource_id:
            logger.warning(
                'Unable to delete the saved search, it does not appear to be '
                'saved in the first place.')
            return False

        resource_url = (
            f'{self.api.api_root}/sketches/{self._sketch.id}/views/'
            f'{self._resource_id}/')
        response = self.api.session.delete(resource_url)
        return error.check_return_status(response, logger)

    @property
    def description(self):
        """Property that returns back the description of the saved search."""
        return self._description

    @description.setter
    def description(self, description):
        """Make changes to the saved search description field."""
        self._description = description
        self.commit()

    @property
    def expected_size(self):
        """Property that returns the expected size of the search query."""
        if self._total_elastic_size:
            return self._total_elastic_size

        self._execute_query(count=True)
        return self._total_elastic_size

    def from_manual(  # pylint: disable=arguments-differ
            self,
            query_string=None,
            query_dsl=None,
            query_filter=None,
            return_fields=None,
            max_entries=None,
            **kwargs):
        """Explore the sketch.

        Args:
            query_string (str): Elasticsearch query string.
            query_dsl (str): Elasticsearch query DSL as JSON string.
            query_filter (dict): Filter for the query as a dict.
            return_fields (str): A comma separated string with a list of fields
                that should be included in the response. Optional and defaults
                to None.
            max_entries (int): Optional integer denoting a best effort to limit
                the output size to the number of events. Events are read in,
                10k at a time so there may be more events in the answer back
                than this number denotes, this is a best effort.
            kwargs (dict[str, object]): Depending on the resource they may
                require different sets of arguments to be able to run a raw
                API request.

        Raises:
            ValueError: if unable to query for the results.
            RuntimeError: if the query is missing needed values, or if the
                sketch is archived.
        """
        super().from_manual(**kwargs)
        if not (query_string or query_filter or query_dsl):
            raise RuntimeError('You need to supply a query')

        self._username = self.api.current_user.username
        self._name = 'From Explore'
        self._description = 'From Explore'

        if query_filter:
            self.query_filter = query_filter

        self._query_string = query_string
        self.query_dsl = query_dsl
        self._return_fields = return_fields

        if max_entries:
            self._max_entries = max_entries

        # TODO: Make use of search templates and aggregations.
        #self._searchtemplate = data.get('searchtemplate', 0)
        #self._aggregations = data.get('aggregation', 0)

        self._created_at = datetime.datetime.now(
            datetime.timezone.utc).isoformat()
        self._updated_at = self._created_at

        self.resource_data = {}

    def from_saved(self, search_id):  # pylint: disable=arguments-differ
        """Initialize the search object from a saved search.

        Args:
            search_id: integer value for the saved
                search (primary key).
        """
        resource_uri = f'sketches/{self._sketch.id}/views/{search_id}/'
        resource_data = self.api.fetch_resource_data(resource_uri)

        data = resource_data.get('objects', [None])[0]
        if not data:
            logger.error('Unable to get any data back from a saved search.')
            return

        label_string = data.get('label_string', '')
        if label_string:
            self._labels = json.loads(label_string)
        else:
            self._labels = []

        self._aggregations = data.get('aggregation', 0)
        self._created_at = data.get('created_at', '')
        self._description = data.get('description', '')
        self._name = data.get('name', '')
        self.query_dsl = data.get('query_dsl', '')
        query_filter = data.get('query_filter', '')
        if query_filter:
            filter_dict = json.loads(query_filter)
            if 'fields' in filter_dict:
                fields = filter_dict.pop('fields')
                return_fields = [x.get('field') for x in fields]
                self.return_fields = ','.join(return_fields)

            self.query_filter = filter_dict
        self._query_string = data.get('query_string', '')
        self._resource_id = search_id
        self._searchtemplate = data.get('searchtemplate', 0)
        self._updated_at = data.get('updated_at', '')
        self._username = data.get('user', {}).get('username', 'System')

        self.resource_data = data

    @property
    def indices(self):
        """Return the current set of indices used in the search."""
        return self._indices

    @indices.setter
    def indices(self, indices):
        """Make changes to the current set of indices."""
        def _is_string_or_int(item):
            return isinstance(item, (str, int))

        if not isinstance(indices, list):
            logger.warning(
                'Indices needs to be a list of strings (indices that were '
                'passed in were not a list).')
            return
        if not all(map(_is_string_or_int, indices)):
            logger.warning(
                'Indices needs to be a list of strings or ints, not all '
                'entries in the indices list are valid string/int.')
            return

        # Indices here can be either a list of timeline names, IDs or a list
        # of search indices. We need to verify that these exist before saving
        # them.
        timelines = {
            t.index_name: t.name for t in self._sketch.list_timelines()}

        valid_ids = [t.id for t in self._sketch.list_timelines()]

        new_indices = []
        for index in indices:
            if index in timelines:
                new_indices.append(index)
                continue

            if isinstance(index, int):
                if index in valid_ids:
                    new_indices.append(str(index))
                    continue

            if index.isdigit():
                if int(index) in valid_ids:
                    new_indices.append(index)
                    continue

            if index in timelines.values():
                new_indices.append(index)

        if not new_indices:
            logger.warning('No valid indices found, not changin the value.')
            return

        self._indices = new_indices

    @property
    def max_entries(self):
        """Return the maximum number of entries in the return value."""
        return self._max_entries

    @max_entries.setter
    def max_entries(self, max_entries):
        """Make changes to the max entries of return values."""
        self._max_entries = max_entries
        if max_entries < self.DEFAULT_SIZE_LIMIT:
            _ = self.query_filter
            self._query_filter['size'] = max_entries
            self._query_filter['terminate_after'] = max_entries
        self.commit()

    @property
    def name(self):
        """Property that returns the query name."""
        return self._name

    @name.setter
    def name(self, name):
        """Make changes to the saved search name."""
        self._name = name
        self.commit()

    def order_ascending(self):
        """Set the order of objects returned back ascending."""
        # Trigger a creation of a query filter if it does not exist.
        _ = self.query_filter
        self._query_filter['order'] = 'asc'

    def order_descending(self):
        """Set the order of objects returned back descending."""
        # Trigger a creation of a query filter if it does not exist.
        _ = self.query_filter
        self._query_filter['order'] = 'desc'

    @property
    def query_dsl(self):
        """Property that returns back the query DSL."""
        return self._query_dsl

    @query_dsl.setter
    def query_dsl(self, query_dsl):
        """Make changes to the query DSL of the search."""
        if query_dsl and isinstance(query_dsl, str):
            query_dsl = json.loads(query_dsl)

        # Special condition of an empty DSL.
        if query_dsl == '""':
            query_dsl = ''

        self._query_dsl = query_dsl
        self.commit()

    @property
    def query_filter(self):
        """Property that returns the query filter."""
        if not self._query_filter:
            self._query_filter = {
                'size': self.DEFAULT_SIZE_LIMIT,
                'terminate_after': self.DEFAULT_SIZE_LIMIT,
                'indices': self.indices,
                'order': 'asc',
                'chips': [],
            }

        query_filter = self._query_filter
        query_filter['chips'] = [x.chip for x in self._chips]
        query_filter['indices'] = self.indices
        return query_filter

    @query_filter.setter
    def query_filter(self, query_filter):
        """Make changes to the query filter."""
        if isinstance(query_filter, str):
            try:
                query_filter = json.loads(query_filter)
            except json.JSONDecodeError as exc:
                raise ValueError('Unable to parse the string as JSON') from exc

        if not isinstance(query_filter, dict):
            raise ValueError('Query filter needs to be a dict.')
        self._query_filter = query_filter
        self._extract_chips(query_filter)
        self.commit()

    @property
    def query_string(self):
        """Property that returns back the query string."""
        return self._query_string

    @query_string.setter
    def query_string(self, query_string):
        """Make changes to the query string of a saved search."""
        self._query_string = query_string
        self.commit()

    def remove_chip(self, chip_index):
        """Remove a chip from the saved search."""
        chip_len = len(self._chips)
        if chip_index > (chip_len + 1):
            raise ValueError(
                f'Unable to remove chip, only {chip_len} chips stored '
                f'(no index {chip_index})')

        try:
            _ = self._chips.pop(chip_index)
        except IndexError as exc:
            raise ValueError(
                f'Unable to remove index {chip_index}, out of range') from exc

        self.commit()

    @property
    def return_fields(self):
        """Property that returns the return_fields."""
        return self._return_fields

    @return_fields.setter
    def return_fields(self, return_fields):
        """Make changes to the return fields."""
        self._return_fields = return_fields
        self.commit()

    @property
    def return_size(self):
        """Return the maximum number of entries in the return value."""
        return self._max_entries

    @return_size.setter
    def return_size(self, return_size):
        """Make changes to the maximum number of entries in the return."""
        self._max_entries = return_size

    def save(self):
        """Save the search in the database.

        Returns:
            String with the identification of the saved search.

        Raises:
            ValueError: if there are values missing in order to save the query.
            RuntimeError: if the search could not be saved.
        """
        if not self.name:
            raise ValueError(
                'No name for the query saved. Please select a name first.')

        if not (self.query_string or self.query_dsl):
            raise ValueError(
                'Need to have either a query DSL or a query string to be '
                'able to save the search.')

        if not self.description:
            logger.warning(
                'No description selected for search, saving without one')

        if self._resource_id:
            resource_url = (
                f'{self.api.api_root}/sketches/{self._sketch.id}/views/'
                f'{self._resource_id}/')
        else:
            resource_url = (
                f'{self.api.api_root}/sketches/{self._sketch.id}/views/')

        query_filter = self.query_filter
        if self.return_fields:
            sketch_data = self._sketch.data
            sketch_meta = sketch_data.get('meta', {})
            mappings = sketch_meta.get('mappings', [])

            use_mappings = []
            for field in self.return_fields.split(','):
                field = field.strip().lower()
                for map_entry in mappings:
                    if map_entry.get('field', '').lower() == field:
                        use_mappings.append(map_entry)
            query_filter['fields'] = use_mappings

        data = {
            'name': self.name,
            'description': self.description,
            'query': self.query_string,
            'filter': query_filter,
            'dsl': self.query_dsl,
            'labels': json.dumps(self.labels),
        }
        response = self.api.session.post(resource_url, json=data)
        status = error.check_return_status(response, logger)
        if not status:
            error.error_message(
                response, 'Unable to save search', error=RuntimeError)

        response_json = error.get_response_json(response, logger)
        search_dict = response_json.get('objects', [{}])[0]
        self._resource_id = search_dict.get('id', 0)
        return f'Saved search to ID: {self._resource_id}'

    def save_as_template(self):
        """Save the search as a search template.

        Returns:
            A search template object (searchtemplate.SearchTemplate).
        """
        if not self._resource_id:
            logger.warning('Search has not been saved first, saving now.')
            return_string = self.save()
            logger.info(return_string)

        template = searchtemplate.SearchTemplate(self.api)
        template.from_search_object(self)

        print(template.save())
        self._searchtemplate = template.id
        return template

    @property
    def scrolling(self):
        """Returns whether scrolling is enabled or not."""
        return self._scrolling

    def scrolling_disable(self):
        """"Disables scrolling."""
        self._scrolling = False

    def scrolling_enable(self):
        """Enable scrolling."""
        self._scrolling = True

    def to_dict(self):
        """Returns a dict with the respone of the query."""
        if not self._raw_response:
            self._execute_query()

        return self._raw_response

    def to_file(self, file_name):
        """Saves the content of the query to a file.

        Args:
            file_name (str): Full path to a file that will store the results
                of the query to as a ZIP file. The ZIP file will contain a
                METADATA file and a CSV with the results from the query.

        Returns:
            Boolean that determines if it was successful.
        """
        old_scrolling = self.scrolling
        self._scrolling = True
        self._execute_query(file_name=file_name)
        self._scrolling = old_scrolling
        return True

    def to_pandas(self):
        """Returns a pandas DataFrame with the response of the query."""
        if not self._raw_response:
            self._execute_query()

        return_list = []
        timelines = {
            t.index_name: t.name for t in self._sketch.list_timelines()}

        return_field_list = []
        return_fields = self._return_fields
        if return_fields:
            if return_fields.startswith('\''):
                return_fields = return_fields[1:]
            if return_fields.endswith('\''):
                return_fields = return_fields[:-1]
            return_field_list = return_fields.split(',')

        for result in self._raw_response.get('objects', []):
            source = result.get('_source', {})
            if not return_fields or '_id' in return_field_list:
                source['_id'] = result.get('_id')
            if not return_fields or '_type' in return_field_list:
                source['_type'] = result.get('_type')
            if not return_fields or '_index' in return_field_list:
                source['_index'] = result.get('_index')
            if not return_fields or '_source' in return_field_list:
                source['_source'] = timelines.get(result.get('_index'))

            return_list.append(source)

        data_frame = pandas.DataFrame(return_list)
        if 'datetime' in data_frame:
            try:
                data_frame['datetime'] = pandas.to_datetime(data_frame.datetime)
            except pandas.errors.OutOfBoundsDatetime:
                pass
        elif 'timestamp' in data_frame:
            try:
                data_frame['datetime'] = pandas.to_datetime(
                    data_frame.timestamp / 1e6, utc=True, unit='s')
            except pandas.errors.OutOfBoundsDatetime:
                pass

        return data_frame

    @property
    def updated_at(self):
        """Property that returns back the updated time of a search."""
        return self._updated_at



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API client library."""
import logging

from . import resource


logger = logging.getLogger('timesketch_api.user')


class User(resource.BaseResource):
    """User object."""

    def __init__(self, api):
        """Initializes the user object."""
        self._object_data = None
        resource_uri = 'users/me/'
        super().__init__(api, resource_uri)

    def _get_data(self):
        """Returns dict from the first object of the resource data."""
        if self._object_data:
            return self._object_data

        data = self.data
        objects = data.get('objects')
        if objects:
            self._object_data = objects[0]
        else:
            self._object_data = {}

        return self._object_data

    @property
    def groups(self):
        """Property that returns the groups the user belongs to."""
        data = self._get_data()
        groups = data.get('groups', [])
        return [x.get('name', '') for x in groups]

    @property
    def is_active(self):
        """Property that returns bool indicating whether the user is active."""
        data = self._get_data()
        return data.get('active', True)

    @property
    def is_admin(self):
        """Property that returns bool indicating whether the user is admin."""
        data = self._get_data()
        return data.get('admin', False)

    @property
    def username(self):
        """Property that returns back the username of the current user."""
        data = self._get_data()
        return data.get('username', 'Unknown')

    def __str__(self):
        """Returns a string representation of the username."""
        user_strings = [self.username]

        if self.is_active:
            user_strings.append('[active]')
        else:
            user_strings.append('[inactive]')

        if self.is_admin:
            user_strings.append('<is admin>')

        return ' '.join(user_strings)



# Copyright 2021 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch API search template object."""

import logging

from . import error
from . import resource
from . import search


logger = logging.getLogger('timesketch_api.searchtemplate')


class SearchTemplate(resource.BaseResource):
    """Search template object. TEST e2e"""

    def __init__(self, api):
        """Initialize the search template object."""
        super().__init__(api, 'searchtemplate/')
        self._description = ''
        self._name = ''
        self._resource_id = None
        self._search_id = None
        self._sketch_id = None

    @property
    def description(self):
        """Property that returns the template description."""
        if self._description:
            return self._description

        if not self._resource_id:
            logger.error('No resource ID, have you loaded the template yet?')
            raise ValueError('Unable to get a name, not loaded yet.')

        data = self.lazyload_data()
        objects = data.get('objects')
        if not objects:
            return 'No description'

        self._description = objects[0].get('description', 'N/A')
        return self._description

    def delete(self):
        """Deletes the saved graph from the store."""
        if not self._resource_id:
            raise ValueError(
                'Unable to delete the search template, since the template '
                'does not seem to be saved, which is required.')

        resource_url = (
            f'{self.api.api_root}/searchtemplate/{self._resource_id}/')
        response = self.api.session.delete(resource_url)
        return error.check_return_status(response, logger)

    def from_saved(self, template_id, sketch_id=0):
        """Initialize the search template from a saved template, by ID value.

        Args:
            template_id: integer value for the saved search template.
            sketch_id: optional integer value for a sketch ID. If not
                provided, an attempt is made to figure it out.

        Raises:
            ValueError: If issues came up during processing.
        """
        self._resource_id = template_id
        self.resource_uri = f'searchtemplate/{self._resource_id}/'

        if sketch_id:
            self._sketch_id = sketch_id
        else:
            data = self.lazyload_data(refresh_cache=True)
            meta = data.get('meta', {})
            sketch_ids = meta.get('sketch_ids', [])
            if len(sketch_ids) > 1:
                sketch_string = ', '.join(sketch_ids)
                raise ValueError(
                    'Search template has too many attached saved searches, '
                    'please pick one from: {0:s}'.format(sketch_string))
            self._sketch_id = sketch_ids[0]

    def from_search_object(self, search_obj):
        """Initialize template from a search object.

        Args:
            search_obj (search.Search): a search object.
        """
        self._search_id = search_obj.id
        self._sketch_id = search_obj.sketch.id

        response = self.api.fetch_resource_data('searchtemplate/')
        meta = response.get('meta', {})
        template_id = 0
        for data in meta.get('collection', []):
            if data.get('search_id') == self._search_id:
                template_id = data.get('template_id', 0)

        if not template_id:
            return

        self._resource_id = template_id
        self.resource_uri = f'searchtemplate/{self._resource_id}/'

    @property
    def id(self):
        """Property that returns back the search template ID."""
        return self._resource_id

    @property
    def name(self):
        """Property that returns the template name."""
        if self._name:
            return self._name

        if not self._resource_id:
            logger.error('No resource ID, have you loaded the template yet?')
            raise ValueError('Unable to get a name, not loaded yet.')

        data = self.lazyload_data()
        objects = data.get('objects')
        if not objects:
            return 'No name'

        self._name = objects[0].get('name', 'N/A')
        return self._name

    def set_sketch(self, sketch=None, sketch_id=None):
        """Set the sketch for the search template.

        Args:
            sketch (sketch.Sketch): an optional sketch object to use as the
                sketch object for the search template.
            sketch_id (int): an optional sketch ID to use as the sketch ID
                for the search template.

        Raises:
            ValueError: If neither a sketch nor a sketch ID is set.
        """
        if not sketch and not sketch_id:
            raise ValueError('Either sketch or sketch ID needs to be set.')

        if sketch:
            self._sketch_id = sketch
        elif isinstance(sketch_id, int):
            self._sketch_id = sketch_id
        else:
            raise ValueError(
                'Sketch needs to be set, or an integer value for '
                'a sketch ID.')

    def save(self):
        """Save the search template."""
        if self._resource_id:
            raise ValueError(
                'The template has already been saved, ATM updates to an '
                'existing template are not yet supported.')

        if not self._search_id:
            raise ValueError(
                'Unable to save the search template since the identification '
                'value of the saved search is not known. The object needs '
                'to be initialized from a previously saved search.')

        data = {
            'search_id': self._search_id,
        }
        resource_url = f'{self.api.api_root}/searchtemplate/'
        response = self.api.session.post(resource_url, json=data)

        status = error.check_return_status(response, logger)
        if not status:
            error.error_message(
                response, 'Unable to save search as a template',
                error=RuntimeError)

        response_json = error.get_response_json(response, logger)
        template_dict = response_json.get('objects', [{}])[0]

        self._resource_id = template_dict.get('id', 0)
        self.resource_uri = f'searchtemplate/{self._resource_id}/'
        return f'Saved search as a template to ID: {self.id}'

    def to_search(self):
        """Returns a search object from a template."""
        if not self._resource_id:
            raise ValueError(
                'Unable to get a search object unless it is tied to a '
                'template.')

        if not self._sketch_id:
            raise ValueError(
                'Unable to get a search object unless it is tied to '
                'a sketch.')

        data = self.lazyload_data(refresh_cache=True)
        objects = data.get('objects')
        if not objects:
            raise ValueError(
                'Unable to get search object, issue with retrieving '
                'template data.')

        template_dict = objects[0]
        sketch = self.api.get_sketch(self._sketch_id)

        search_obj = search.Search(sketch=sketch)
        search_obj.from_manual(
            query_string=template_dict.get('query_string'),
            query_dsl=template_dict.get('query_dsl'),
            query_filter=template_dict.get('query_filter'))
        search_obj.name = template_dict.get('name', 'No Name')
        search_obj.description = template_dict.get(
            'description', 'No Description')
        return search_obj



# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Timesketch API client"""
from __future__ import unicode_literals

import unittest
import mock

from . import client
from . import sketch as sketch_lib
from . import test_lib


class TimesketchApiTest(unittest.TestCase):
    """Test TimesketchApi"""

    @mock.patch('requests.Session', test_lib.mock_session)
    def setUp(self):
        """Setup test case."""
        self.api_client = client.TimesketchApi(
            'http://127.0.0.1', 'test', 'test')

    def test_fetch_resource_data(self):
        """Test fetch resource."""
        response = self.api_client.fetch_resource_data('sketches/')
        self.assertIsInstance(response, dict)

    # TODO: Add test for create_sketch()

    def test_get_sketch(self):
        """Test to get a sketch."""
        sketch = self.api_client.get_sketch(1)
        self.assertIsInstance(sketch, sketch_lib.Sketch)
        self.assertEqual(sketch.id, 1)
        self.assertEqual(sketch.name, 'test')
        self.assertEqual(sketch.description, 'test')

    def test_get_sketches(self):
        """Test to get a list of sketches."""
        sketches = list(self.api_client.list_sketches())
        self.assertIsInstance(sketches, list)
        self.assertEqual(len(sketches), 1)
        self.assertIsInstance(sketches[0], sketch_lib.Sketch)
